import datetime

import pytest
import pytz

from xocto import settlement_periods


UTC_TZ = datetime.timezone.utc
GB_TZ = pytz.timezone("Europe/London")


@pytest.mark.parametrize(
    "sp,date,expected",
    [
        # British time is GMT
        (
            1,
            datetime.date(2017, 1, 1),
            datetime.datetime(2016, 12, 31, hour=23, tzinfo=UTC_TZ),
        ),
        (
            48,
            datetime.date(2017, 1, 1),
            datetime.datetime(2017, 1, 1, hour=22, minute=30, tzinfo=UTC_TZ),
        ),
        # British time is GMT
        (
            1,
            datetime.date(2017, 7, 1),
            datetime.datetime(2017, 6, 30, hour=22, tzinfo=UTC_TZ),
        ),
        (
            48,
            datetime.date(2017, 7, 1),
            datetime.datetime(2017, 7, 1, hour=21, minute=30, tzinfo=UTC_TZ),
        ),
        # Change clock forward day, from GMT to BST
        (
            3,
            datetime.date(2017, 3, 26),
            datetime.datetime(2017, 3, 26, hour=0, tzinfo=UTC_TZ),
        ),
        (
            5,
            datetime.date(2017, 3, 26),
            datetime.datetime(2017, 3, 26, hour=1, tzinfo=UTC_TZ),
        ),
        (
            7,
            datetime.date(2017, 3, 26),
            datetime.datetime(2017, 3, 26, hour=2, tzinfo=UTC_TZ),
        ),
        (
            45,
            datetime.date(2017, 3, 26),
            datetime.datetime(2017, 3, 26, hour=21, tzinfo=UTC_TZ),
        ),
        # Change clock backward day, from BST to GMT
        (
            5,
            datetime.date(2017, 10, 29),
            datetime.datetime(2017, 10, 29, hour=0, tzinfo=UTC_TZ),
        ),
        (
            7,
            datetime.date(2017, 10, 29),
            datetime.datetime(2017, 10, 29, hour=1, tzinfo=UTC_TZ),
        ),
        (
            9,
            datetime.date(2017, 10, 29),
            datetime.datetime(2017, 10, 29, hour=2, tzinfo=UTC_TZ),
        ),
        (
            49,
            datetime.date(2017, 10, 29),
            datetime.datetime(2017, 10, 29, hour=22, tzinfo=UTC_TZ),
        ),
    ],
)
def test_convert_sp_and_date_to_utc_for_wholesale(sp, date, expected):
    """
    Test the convert_sp_and_date_to_utc function within a wholesale context
    for days where british time is the same as GMT, where british time is BST, and change days
    """
    assert (
        settlement_periods.convert_sp_and_date_to_utc(sp, date, is_wholesale=True)
        == expected
    )


@pytest.mark.parametrize(
    "utc,sp,date",
    [
        # British time is GMT
        (
            datetime.datetime(2016, 12, 31, hour=23, minute=30, tzinfo=UTC_TZ),
            2,
            datetime.date(2017, 1, 1),
        ),
        (
            datetime.datetime(2017, 1, 1, hour=22, tzinfo=UTC_TZ),
            47,
            datetime.date(2017, 1, 1),
        ),
        # British time is GMT
        (
            datetime.datetime(2017, 6, 30, hour=22, minute=30, tzinfo=UTC_TZ),
            2,
            datetime.date(2017, 7, 1),
        ),
        (
            datetime.datetime(2017, 7, 1, hour=21, tzinfo=UTC_TZ),
            47,
            datetime.date(2017, 7, 1),
        ),
        # Change clock forward day, from GMT to BST
        (
            datetime.datetime(2017, 3, 26, hour=0, minute=30, tzinfo=UTC_TZ),
            4,
            datetime.date(2017, 3, 26),
        ),
        (
            datetime.datetime(2017, 3, 26, hour=1, minute=30, tzinfo=UTC_TZ),
            6,
            datetime.date(2017, 3, 26),
        ),
        (
            datetime.datetime(2017, 3, 26, hour=2, minute=30, tzinfo=UTC_TZ),
            8,
            datetime.date(2017, 3, 26),
        ),
        (
            datetime.datetime(2017, 3, 26, hour=21, minute=30, tzinfo=UTC_TZ),
            46,
            datetime.date(2017, 3, 26),
        ),
        # Change clock backward day, from BST to GMT
        (
            datetime.datetime(2017, 10, 29, hour=0, minute=30, tzinfo=UTC_TZ),
            6,
            datetime.date(2017, 10, 29),
        ),
        (
            datetime.datetime(2017, 10, 29, hour=1, minute=30, tzinfo=UTC_TZ),
            8,
            datetime.date(2017, 10, 29),
        ),
        (
            datetime.datetime(2017, 10, 29, hour=2, minute=30, tzinfo=UTC_TZ),
            10,
            datetime.date(2017, 10, 29),
        ),
        (
            datetime.datetime(2017, 10, 29, hour=22, minute=30, tzinfo=UTC_TZ),
            50,
            datetime.date(2017, 10, 29),
        ),
    ],
)
def test_convert_utc_to_sp_and_date_for_wholesale(utc, sp, date):
    """
    Test the convert_utc_to_sp_and_date function within a wholesale context
    for days where british time is the same as GMT, where british time is BST, and change days
    """
    assert settlement_periods.convert_utc_to_sp_and_date(utc, is_wholesale=True) == (
        sp,
        date,
    )


@pytest.mark.parametrize(
    "sp,date,expected",
    [
        # British time is GMT
        (
            1,
            datetime.date(2016, 1, 1),
            datetime.datetime(2016, 1, 1, hour=0, tzinfo=UTC_TZ),
        ),
        (
            4,
            datetime.date(2016, 1, 1),
            datetime.datetime(2016, 1, 1, hour=1, minute=30, tzinfo=UTC_TZ),
        ),
        (
            20,
            datetime.date(2016, 1, 1),
            datetime.datetime(2016, 1, 1, hour=9, minute=30, tzinfo=UTC_TZ),
        ),
        # British time is BST
        (
            1,
            datetime.date(2016, 7, 1),
            datetime.datetime(2016, 6, 30, hour=23, tzinfo=UTC_TZ),
        ),
        (
            4,
            datetime.date(2016, 7, 1),
            datetime.datetime(2016, 7, 1, hour=0, minute=30, tzinfo=UTC_TZ),
        ),
        (
            20,
            datetime.date(2016, 7, 1),
            datetime.datetime(2016, 7, 1, hour=8, minute=30, tzinfo=UTC_TZ),
        ),
        # Change clock forward day, from GMT to BST
        (
            1,
            datetime.date(2016, 3, 27),
            datetime.datetime(2016, 3, 27, hour=0, tzinfo=UTC_TZ),
        ),
        (
            4,
            datetime.date(2016, 3, 27),
            datetime.datetime(2016, 3, 27, hour=1, minute=30, tzinfo=UTC_TZ),
        ),
        (
            20,
            datetime.date(2016, 3, 27),
            datetime.datetime(2016, 3, 27, hour=9, minute=30, tzinfo=UTC_TZ),
        ),
        # Change clock backward day, from BST to GMT
        (
            1,
            datetime.date(2015, 10, 25),
            datetime.datetime(2015, 10, 24, hour=23, tzinfo=UTC_TZ),
        ),
        (
            4,
            datetime.date(2015, 10, 25),
            datetime.datetime(2015, 10, 25, hour=0, minute=30, tzinfo=UTC_TZ),
        ),
        (
            20,
            datetime.date(2015, 10, 25),
            datetime.datetime(2015, 10, 25, hour=8, minute=30, tzinfo=UTC_TZ),
        ),
    ],
)
def test_convert_sp_and_date_to_utc_for_retail(sp, date, expected):
    """
    Test the convert_sp_and_date_to_utc function within a retail context
    for days where british time is the same as GMT, where british time is BST, and change days
    """
    assert settlement_periods.convert_sp_and_date_to_utc(sp, date) == expected


@pytest.mark.parametrize(
    "utc,sp,date",
    [
        # British time is GMT
        (
            datetime.datetime(2016, 1, 1, hour=0, tzinfo=UTC_TZ),
            1,
            datetime.date(2016, 1, 1),
        ),
        (
            datetime.datetime(2016, 1, 1, hour=1, minute=30, tzinfo=UTC_TZ),
            4,
            datetime.date(2016, 1, 1),
        ),
        (
            datetime.datetime(2016, 1, 1, hour=9, minute=30, tzinfo=UTC_TZ),
            20,
            datetime.date(2016, 1, 1),
        ),
        # British time is BST
        (
            datetime.datetime(2016, 6, 30, hour=23, tzinfo=UTC_TZ),
            1,
            datetime.date(2016, 7, 1),
        ),
        (
            datetime.datetime(2016, 7, 1, hour=0, minute=30, tzinfo=UTC_TZ),
            4,
            datetime.date(2016, 7, 1),
        ),
        (
            datetime.datetime(2016, 7, 1, hour=8, minute=30, tzinfo=UTC_TZ),
            20,
            datetime.date(2016, 7, 1),
        ),
        # Change clock forward day, From GMT to BST
        (
            datetime.datetime(2016, 3, 27, hour=0, tzinfo=UTC_TZ),
            1,
            datetime.date(2016, 3, 27),
        ),
        (
            datetime.datetime(2016, 3, 27, hour=1, minute=30, tzinfo=UTC_TZ),
            4,
            datetime.date(2016, 3, 27),
        ),
        (
            datetime.datetime(2016, 3, 27, hour=9, minute=30, tzinfo=UTC_TZ),
            20,
            datetime.date(2016, 3, 27),
        ),
        # Change clock backward day, from BST to GMT
        (
            datetime.datetime(2016, 10, 24, hour=23, tzinfo=UTC_TZ),
            1,
            datetime.date(2016, 10, 25),
        ),
        (
            datetime.datetime(2016, 10, 25, hour=0, minute=30, tzinfo=UTC_TZ),
            4,
            datetime.date(2016, 10, 25),
        ),
        (
            datetime.datetime(2016, 10, 25, hour=8, minute=30, tzinfo=UTC_TZ),
            20,
            datetime.date(2016, 10, 25),
        ),
    ],
)
def test_convert_utc_to_sp_and_date_for_retail(utc, sp, date):
    """
    Test the convert_utc_to_sp_and_date function within a retail context
    for days where british time is the same as GMT, where british time is BST, and change days
    """
    assert settlement_periods.convert_utc_to_sp_and_date(utc) == (sp, date)


@pytest.mark.parametrize(
    "start,end,periods",
    [
        (datetime.datetime(2016, 1, 1, 0, 0), datetime.datetime(2016, 1, 1, 0, 30), 1),
        (datetime.datetime(2016, 1, 1, 0, 0), datetime.datetime(2016, 1, 1, 4, 0), 8),
        (datetime.datetime(2016, 1, 1, 0, 0), datetime.datetime(2016, 1, 2, 0, 30), 49),
        # Clocks go forward here: only 23 hours in the day
        (
            datetime.datetime(2018, 3, 25, 0, 0),
            datetime.datetime(2018, 3, 26, 0, 0),
            46,
        ),
        # Clocks go backwards here: 25 hours in the day
        (
            datetime.datetime(2018, 10, 28, 0, 0),
            datetime.datetime(2018, 10, 29, 0, 0),
            50,
        ),
    ],
)
def test_number_of_settlement_periods_in_timedelta(start, end, periods):
    start = GB_TZ.normalize(GB_TZ.localize(start))
    end = GB_TZ.normalize(GB_TZ.localize(end))
    delta = end - start

    assert settlement_periods.number_of_periods_in_timedelta(delta) == periods


@pytest.mark.parametrize(
    "time,is_dst,expected",
    [
        (
            datetime.datetime(2016, 1, 1, 0, 0),
            None,
            datetime.datetime(2016, 1, 1, 0, 0),
        ),
        (
            datetime.datetime(2016, 1, 1, 0, 30),
            None,
            datetime.datetime(2016, 1, 1, 0, 30),
        ),
        (
            datetime.datetime(2016, 1, 1, 0, 15),
            None,
            datetime.datetime(2016, 1, 1, 0, 0),
        ),
        (
            datetime.datetime(2016, 1, 1, 0, 45),
            None,
            datetime.datetime(2016, 1, 1, 0, 30),
        ),
        # Hour before clocks change forward
        (
            datetime.datetime(2016, 3, 27, 0, 0),
            None,
            datetime.datetime(2016, 3, 27, 0, 0),
        ),
        # Hour after clocks change forward
        (
            datetime.datetime(2016, 3, 27, 2, 0),
            None,
            datetime.datetime(2016, 3, 27, 2, 0),
        ),
        # Hour before clocks change back
        (
            datetime.datetime(2016, 10, 30, 1, 0),
            True,
            datetime.datetime(2016, 10, 30, 1, 0),
        ),
        # Hour after clocks change back
        (
            datetime.datetime(2016, 10, 30, 1, 0),
            False,
            datetime.datetime(2016, 10, 30, 1, 0),
        ),
    ],
)
def test_round_down_local_time(time, is_dst, expected):
    # Operations within the hour stay in the same timezone & DST
    local_time = GB_TZ.localize(time, is_dst=is_dst)

    expected = GB_TZ.localize(expected, is_dst=is_dst)

    assert settlement_periods._round_local_down_to_hh(local_time) == expected



from decimal import ROUND_DOWN, ROUND_HALF_DOWN, ROUND_HALF_UP, ROUND_UP
from decimal import Decimal as D

import pytest
from hypothesis import assume, given, strategies

from xocto import numbers


@pytest.mark.parametrize(
    "number_to_round, base, rounding_strategy, expected_result",
    [
        # DEFAULT ROUNDING STRATEGY (round-half-even)
        (153, 5, None, 155),
        # 23 * 12,793 == 294,239, which is the closest multiple of 23 to our starting number.
        (294_242, 23, None, 294_239),
        (64, 1, None, 64),
        # Numbers halfway between the rounding boundary will always round down
        (15, 30, None, 0),
        # CUSTOM ROUNDING STRATEGIES
        (153, 5, ROUND_DOWN, 150),
        (D("152.5"), 5, ROUND_HALF_DOWN, 150),
        (D("152.5"), 5, ROUND_HALF_UP, 155),
        # Check that rounding doesn't round beyond a quantised value
        (155, 5, ROUND_DOWN, 155),
        (150, 5, ROUND_UP, 150),
    ],
)
def test_quantise(number_to_round, base, rounding_strategy, expected_result):
    args = (number_to_round, base)
    kwargs = dict(rounding=rounding_strategy) if rounding_strategy else {}
    assert numbers.quantise(*args, **kwargs) == expected_result


def test_truncate_decimal_places():
    assert numbers.truncate_decimal_places(D("123.45"), 1) == 123.4
    assert numbers.truncate_decimal_places(D("123.456"), 1) == 123.4
    assert numbers.truncate_decimal_places(D("123.4"), 2) == 123.40
    assert numbers.truncate_decimal_places(D("123.45"), 0) == 123.0


class TestClipToRange:
    @given(strategies.integers(), strategies.integers(), strategies.integers())
    def test_always_clips_ints_to_range(self, val, minval, maxval):
        assume(minval <= maxval)
        clipped_val = numbers.clip_to_range(val, minval=minval, maxval=maxval)
        assert minval <= clipped_val <= maxval

    @given(strategies.dates(), strategies.dates(), strategies.dates())
    def test_always_clips_dates_to_range(self, val, minval, maxval):
        assume(minval <= maxval)
        clipped_val = numbers.clip_to_range(val, minval=minval, maxval=maxval)
        assert minval <= clipped_val <= maxval

    @given(
        strategies.floats(allow_nan=False),
        strategies.floats(allow_nan=False),
        strategies.floats(allow_nan=False),
    )
    def test_always_clips_floats_to_range(self, val, minval, maxval):
        assume(minval <= maxval)
        clipped_val = numbers.clip_to_range(val, minval=minval, maxval=maxval)
        assert minval <= clipped_val <= maxval

    def test_raises_if_min_gt_max(self):
        with pytest.raises(AssertionError):
            numbers.clip_to_range(5, minval=8, maxval=4)


class TestRandomInt:
    @pytest.mark.parametrize("length", [2, 10, 20])
    def test_correct_length_returned(self, length):
        number = numbers.random_int(length=length)
        assert len(str(number)) == length

    @pytest.mark.parametrize("length", [-1, 0, 1])
    def test_too_small_length_raises_exception(self, length):
        with pytest.raises(ValueError):
            numbers.random_int(length=length)



from __future__ import annotations

import ddtrace


tracer = ddtrace.tracer
wrap = tracer.wrap


def set_tags(tags: dict[str | bytes, str]) -> None:
    """
    Set multiple tags on the current span.
    """
    span = tracer.current_span()
    if span:
        span.set_tags(tags)


def set_tag(key: str, value: object) -> None:
    """
    Set a tag on the current span.
    """
    span = tracer.current_span()
    if span:
        span.set_tag(key, value)


def set_global_tag(key: str, value: object) -> None:
    """
    Set a tag on the current root span.

    These tags will be associated to the entire Datadog trace.
    """
    span = tracer.current_root_span()
    if span:
        span.set_tag(key, value)



from __future__ import annotations

import datetime
import enum
import functools
import itertools
import operator
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
)

from dateutil import relativedelta

from xocto.types import generic

from . import localtime


class RangeBoundaries(enum.Enum):
    EXCLUSIVE_EXCLUSIVE = "()"
    EXCLUSIVE_INCLUSIVE = "(]"
    INCLUSIVE_EXCLUSIVE = "[)"
    INCLUSIVE_INCLUSIVE = "[]"

    @classmethod
    def from_bounds(
        cls, left_exclusive: bool, right_exclusive: bool
    ) -> "RangeBoundaries":
        """
        Convenience method to get the relevant boundary type by specifiying the exclusivity of each
        end.
        """
        return {
            (True, True): RangeBoundaries.EXCLUSIVE_EXCLUSIVE,
            (True, False): RangeBoundaries.EXCLUSIVE_INCLUSIVE,
            (False, True): RangeBoundaries.INCLUSIVE_EXCLUSIVE,
            (False, False): RangeBoundaries.INCLUSIVE_INCLUSIVE,
        }[(left_exclusive, right_exclusive)]


T = TypeVar("T", bound=generic.Comparable)  # type: ignore[type-arg]


@functools.total_ordering
class Range(Generic[T]):
    """
    The basic concept of an range of some sort of comparable item, specified by its endpoints
    and boundaries (whether the endpoints are inclusive or exclusive). This class provides some
    useful helpers for working with ranges.

    Usage:

    * Ranges are declared by specifying their endpoints and boundaries. If the boundaries is
      omitted, then they are inclusive-exclusive by default

        >>> Range(0, 2, boundaries=RangeBoundaries.EXCLUSIVE_INCLUSIVE)
        <Range: (0,2]>
        >>> Range(0, 2, boundaries="[]")
        <Range: [0,2]>
        >>> r = Range(0, 2)
        >>> print(f"{r}")
        "[0,2)"

    * If an endpoint is set as None, then that means that the range is effectively infinite.
      Infinite ranges must have exclusive bounds for the infinite ends. We provide the continuum
      to the continnum
      helper to get an unbounded range.

        >>> Range(0, None)
        <Range: [0, None)
        >>> Range(0, None, boundaries="[]")  # Invalid
        >>> Range.continuum()  # Helper function
        <Range: (None, None)>

    * Ranges can be declared for any comparable type

        >>> int_erval: Range[int] = Range(0, 2)
        >>> date_erval: Range[date] = Range(date(2020, 1, 1), date(2020, 6, 6))
        >>> string_erval: Range[str] = Range("ardvark", "zebra")  # Uses lexical ordering

    * Ranges are themselves comparable. Two ranges are ordered by their start, with their end used
      to break ties

        >>> sorted([Range(1, 4), Range(0, 5)])
        [<Range: [0,5)>, <Range: [1,4)>]
        >>> sorted([Range(1, 2), Range(None, 2)])
        [<Range: [None,2)>, <Range: [1,2)>]
        >>> sorted([Range(3, 5), Range(3, 4)])
        [<Range: [3,4)>, <Range: [4,5)>]
        >>> sorted([Range(0, 2, boundaries=b) for b in RangeBoundaries])
        [<Range: [0,2)>, <Range: [0,2]>, <Range: (0,2)>, <Range: (0,2]>]

    * The `in` operator is provided to check if an item of the range type is within the
      the range

        >>> 0 in Range(0, 2)
        True
        >>> 2 in Range(0, 2)
        False
        >>> date(2020, 1, 1) in Range(date(2020, 1, 2), date(2020, 1, 5))
        False

    * The `intersection` function (which is aliased to the and (&) operator) will return the
      overlap of two ranges, or None if they are disjoint
        >>> Range(0, 2).intersection(Range(1, 4))
        <Range: [1,2)>
        >>> Range(1, 2) & Range(3, 4)
        None

    * The `is_disjoint` function will tell you if two ranges are disjoint
        >>> Range(0, 2).is_disjoint(Range(3, 5))
        True
        >>> Range(0, 2).is_disjoint(Range(2, 5))  # Since the default is not right-inclusive
        True

    * The `union` function (which is aliased to the or (|) operator), will return a range which
      covers two overlapping (or touching) ranges, or None if they are disjoint.
        >>> Range(0, 2).union(Range(1, 3))
        <Range: [0,3)>
        >>> Range(0, 2) | Range(2, 4)
        <Range: [0,4)>
        >>> Range(0, 2) | Range(3, 4)
        None
        >>> Range(0, 2) | Range(2, 4, boundaries="(]")
        None
        >>> Range(0, 2, boundaries="[]") | Range(3, 4)  # Since Range doesn't understand that 2 and 3 are adjacent.
        None

    * The `difference` function (which is aliased to the subtraction (-) operator), will return a
      range which contains the bit of this range which is not covered by the other range, or a
      rangeset which contains the bits of this range which are not covered (or None if the other
      range covers this one).
        >>> Range(0, 4) - Range(2, 4)
        <Range: [0,2)>
        >>> Range(0, 4) - Range(2, 3)
        <RangeSet: {[0,2), [3,4)}>
        >>> Range(0, 4) - Range(0, 5)
        None
    """

    __slots__ = (
        "start",
        "end",
        "boundaries",
        "_is_left_exclusive",
        "_is_left_inclusive",
        "_is_right_exclusive",
        "_is_right_inclusive",
    )

    start: Optional[T]
    end: Optional[T]
    boundaries: RangeBoundaries
    _is_left_exclusive: bool
    _is_left_inclusive: bool
    _is_right_exclusive: bool
    _is_right_inclusive: bool

    def __init__(
        self,
        start: Optional[T],
        end: Optional[T],
        *,
        boundaries: Union[str, RangeBoundaries] = RangeBoundaries.INCLUSIVE_EXCLUSIVE,
    ):
        """
        Validate that the provided start and end values create a valid range for the boundaries.

        Also set some convenience properties for internal use.
        """
        # Make sure that we are dealing with an enum in case the class was constructed with the
        # string representation of the boundaries
        range_boundaries: RangeBoundaries = RangeBoundaries(boundaries)

        _is_left_exclusive = range_boundaries in [
            RangeBoundaries.EXCLUSIVE_EXCLUSIVE,
            RangeBoundaries.EXCLUSIVE_INCLUSIVE,
        ]
        _is_right_exclusive = range_boundaries in [
            RangeBoundaries.EXCLUSIVE_EXCLUSIVE,
            RangeBoundaries.INCLUSIVE_EXCLUSIVE,
        ]
        _is_left_inclusive = not _is_left_exclusive
        _is_right_inclusive = not _is_right_exclusive

        if start is None:
            if _is_left_inclusive:
                raise ValueError("Range with unbounded start must be left-exclusive")
        if end is None:
            if _is_right_inclusive:
                raise ValueError("Range with unbounded end must be right-exclusive")
        elif start is not None:
            check_op: Callable[[Any, Any], bool] = {
                RangeBoundaries.EXCLUSIVE_EXCLUSIVE: operator.lt,
                RangeBoundaries.EXCLUSIVE_INCLUSIVE: operator.lt,
                RangeBoundaries.INCLUSIVE_EXCLUSIVE: operator.lt,
                RangeBoundaries.INCLUSIVE_INCLUSIVE: operator.le,
            }[range_boundaries]
            if not check_op(start, end):
                raise ValueError("Invalid boundaries for range")

        object.__setattr__(self, "start", start)
        object.__setattr__(self, "end", end)
        object.__setattr__(self, "boundaries", range_boundaries)
        object.__setattr__(self, "_is_left_exclusive", _is_left_exclusive)
        object.__setattr__(self, "_is_left_inclusive", _is_left_inclusive)
        object.__setattr__(self, "_is_right_exclusive", _is_right_exclusive)
        object.__setattr__(self, "_is_right_inclusive", _is_right_inclusive)

    @classmethod
    def continuum(cls) -> Range[T]:
        """
        Return a range representing the continnum.
        """
        return cls(None, None, boundaries="()")

    def __str__(self) -> str:
        if self._is_left_exclusive:
            lbracket = "("
        else:
            lbracket = "["

        if self._is_right_exclusive:
            rbracket = ")"
        else:
            rbracket = "]"

        return f"{lbracket}{self.start},{self.end}{rbracket}"

    def __repr__(self) -> str:
        return f"<Range: {str(self)}>"

    def __setattr__(self, name: str, value: Any) -> None:
        if name in type(self).__slots__:
            raise AttributeError("Can't set attributes")
        else:
            super().__setattr__(name, value)

    def __hash__(self) -> int:
        return hash((self.start, self.end, self.boundaries))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Range):
            return False

        return (self.start, self.end, self.boundaries) == (
            other.start,
            other.end,
            other.boundaries,
        )

    def __lt__(self, other: "Range[T]") -> bool:
        if self.start == other.start:
            if self._is_left_exclusive and not other._is_left_exclusive:
                return False

            if (not self._is_left_exclusive) and other._is_left_exclusive:
                return True

            if self.end == other.end:
                if self._is_right_exclusive and not other._is_right_exclusive:
                    return True

                return False
            else:
                # If one endpoint is None then that range is greater, otherwise compare them
                return (other.end is None) or (
                    self.end is not None and self.end < other.end
                )
        else:
            # If one endpoint is None then that range is lesser, otherwise compare them
            return (self.start is None) or (
                other.start is not None and self.start < other.start
            )

    def __contains__(self, item: T) -> bool:
        """
        Check if the provided item is inside the range.
        """
        return self._is_inside_left_bound(item) and self._is_inside_right_bound(item)

    def __copy__(self) -> Range[T]:
        """
        Return self.

        Ranges are immutable, so there is no need to create a copy.
        """
        return self

    def __deepcopy__(self, memo: dict[Any, Any]) -> Range[T]:
        """
        Return self.

        Ranges are immutable, so there is no need to create a copy.
        """
        return self

    def _is_inside_left_bound(self, item: T) -> bool:
        """
        Check if the provided item is inside our left bound.
        """
        if self.start is None:
            return True
        elif self._is_left_exclusive:
            return item > self.start
        else:
            return item >= self.start

    def _is_inside_right_bound(self, item: T) -> bool:
        """
        Check if the provided item is inside our right bound.
        """
        if self.end is None:
            return True
        elif self._is_right_exclusive:
            return item < self.end
        else:
            return item <= self.end

    def is_disjoint(self, other: "Range[T]") -> bool:
        """
        Test whether the two ranges are disjoint.
        """
        if self.end is not None and other.start is not None:
            if not (
                self._is_inside_right_bound(other.start)
                and other._is_inside_left_bound(self.end)
            ):
                return True

        if self.start is not None and other.end is not None:
            if not (
                self._is_inside_left_bound(other.end)
                and other._is_inside_right_bound(self.start)
            ):
                return True

        return False

    def intersection(self, other: "Range[T]") -> Optional["Range[T]"]:
        """
        Return the intersection of the two ranges if it exists, or None if they are disjoint.
        """
        if self.is_disjoint(other):
            return None

        range_l, range_r = (self, other) if self < other else (other, self)

        # Since we have sorted the two ranges, the left will always be determined by range_r, but
        # the right depends on whether range_l contains range_r or not.
        #
        # This has the effect of making the intersection prefer an inclusive right boundary to an
        # equivalent exclusive one (e.g. [0,2] is preferred over [0,3))

        if range_l.end is not None and range_r._is_inside_right_bound(range_l.end):
            end: Optional[T] = range_l.end
            right_exclusive = range_l._is_right_exclusive
        else:
            end = range_r.end
            right_exclusive = range_r._is_right_exclusive

        boundaries = RangeBoundaries.from_bounds(
            range_r._is_left_exclusive, right_exclusive
        )

        return Range(range_r.start, end, boundaries=boundaries)

    def union(self, other: "Range[T]") -> Optional["Range[T]"]:
        """
        If two ranges overlap (or are adjacent), return an range covering the two ranges. If the
        two ranges are disjoint, return None.
        """
        range_l, range_r = (self, other) if self < other else (other, self)

        if range_l.is_disjoint(range_r) and not (
            range_l.end == range_r.start
            and (range_l._is_right_inclusive or range_r._is_left_inclusive)
        ):
            return None

        # Since we have sorted the two ranges, the left will always be determined by range_l, but
        # the right depends on whether range_l contains range_r or not.
        #
        # This has the effect of making the resulting union prefer an exclusive right boundary to
        # the equivalent inclusive one (e.g. [0,3) is preferred over [0,2]).
        if range_r.end is not None and range_l._is_inside_right_bound(range_r.end):
            end = range_l.end
            right_exclusive = range_l._is_right_exclusive
        else:
            end = range_r.end
            right_exclusive = range_r._is_right_exclusive

        boundaries = RangeBoundaries.from_bounds(
            range_l._is_left_exclusive, right_exclusive
        )

        return Range(range_l.start, end, boundaries=boundaries)

    def difference(
        self, other: "Range[T]"
    ) -> Optional[Union["Range[T]", "RangeSet[T]"]]:
        """
        Return a range or rangeset consisting of the bits of this range that do not intersect the
        other range (or None if this range is covered by the other range).
        """
        if self.is_disjoint(other):
            return self

        if (self.start is None and other.start is not None) or (
            self.start is not None
            and other.start is not None
            and not other._is_inside_left_bound(self.start)
            and self._is_inside_left_bound(other.start)
        ):
            boundaries = RangeBoundaries.from_bounds(
                self._is_left_exclusive, other._is_left_inclusive
            )
            lower_part: Optional["Range[T]"] = Range(
                self.start, other.start, boundaries=boundaries
            )
        else:
            lower_part = None

        if (self.end is None and other.end is not None) or (
            self.end is not None
            and other.end is not None
            and not other._is_inside_right_bound(self.end)
            and self._is_inside_right_bound(other.end)
        ):
            boundaries = RangeBoundaries.from_bounds(
                other._is_right_inclusive, self._is_right_exclusive
            )
            upper_part: Optional["Range[T]"] = Range(
                other.end, self.end, boundaries=boundaries
            )
        else:
            upper_part = None

        if lower_part is None and upper_part is None:
            return None
        elif lower_part is not None and upper_part is not None:
            return RangeSet([lower_part, upper_part])
        else:
            return lower_part or upper_part

    def __and__(self, other: "Range[T]") -> Optional["Range[T]"]:
        """
        Wrapper for intersection to allow the & operator to be used.
        """
        return self.intersection(other)

    def __or__(self, other: "Range[T]") -> Optional["Range[T]"]:
        """
        Wrapper for union to allow the | operator to be used.
        """
        return self.union(other)

    def __sub__(self, other: "Range[T]") -> Optional[Union["Range[T]", "RangeSet[T]"]]:
        """
        Wrapper for difference to allow the - operator to be used.
        """
        return self.difference(other)

    def is_left_finite(self) -> bool:
        return self.start is not None

    def is_right_finite(self) -> bool:
        return self.end is not None

    def is_finite(self) -> bool:
        return self.is_left_finite() and self.is_right_finite()


class FiniteRange(Range[T]):
    """
    A FiniteRange represents a range that MUST have finite endpoints (i.e. they cannot be None).

    This is mostly to add type checking, as we can specify that functions require a finite range
    and then skip checking if the endpoints are None.
    """

    __slots__ = ()

    # Redefine types in base class
    start: T
    end: T

    def intersection(self, other: Range[T]) -> Optional["FiniteRange[T]"]:
        """
        Intersections with finite ranges will always be finite.
        """
        return cast("FiniteRange[T]", super().intersection(other))

    def __and__(self, other: Range[T]) -> Optional["FiniteRange[T]"]:
        return self.intersection(other)


class HalfFiniteRange(Range[T]):
    """
    This is also for type-checking, but represents a very common range type in Kraken (possibly
    the most common).

    Specifically, ranges that MUST have a finite inclusive left endpoint and a (possibly infinite)
    exclusive right endpoint.

    However LeftFiniteInclusiveRightExclusiveRange is a bit of a mouthful so we're going with
    HalfFiniteRange.
    """

    __slots__ = ()

    # Redefine types in base class
    start: T

    def __init__(self, start: T, end: Optional[T] = None):
        super().__init__(start, end, boundaries=RangeBoundaries.INCLUSIVE_EXCLUSIVE)

    def intersection(self, other: Range[T]) -> Optional["HalfFiniteRange[T]"]:
        """
        Intersections with half finite ranges will always be half finite.
        """
        return cast("HalfFiniteRange[T]", super().intersection(other))

    def __and__(self, other: Range[T]) -> Optional["HalfFiniteRange[T]"]:
        return self.intersection(other)


# Type aliases for common range types
DatetimeRange = Range[datetime.datetime]
HalfFiniteDatetimeRange = HalfFiniteRange[datetime.datetime]


class RangeSet(Generic[T]):
    """
    A RangeSet represents an ordered set of disjoint ranges. It is constructed from an iterable of
    Ranges (which can be omited to create an empty RangeSet):

        >>> RangeSet()  # empty RangeSet
        <RangeSet: {}>
        >>> rs = RangeSet([Range(0, 1), Range(2, 4)])  # Single iterable of ranges
        >>> print(f"{rs}")
        "{[0,1), [2, 4)}"

    Overlapping Ranges are condensed when they are added to a set:

        >>> RangeSet([Range(0, 3), Range(2, 4)]) == RangeSet([Range(0, 4)])
        True
    """

    def __init__(self, source_ranges: Optional[Iterable[Range[T]]] = None):
        if source_ranges is None:
            # We need to type this as a Sequence so that it is covariant
            self._ranges: Sequence[Range[T]] = []
        else:
            self._ranges = self._condense_range_list(source_ranges)

    def __str__(self) -> str:
        ranges = ", ".join(str(r) for r in self._ranges)
        return f"{{{ranges}}}"

    def __repr__(self) -> str:
        return f"<RangeSet: {str(self)}>"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, RangeSet):
            return False

        return self._ranges == other._ranges

    def add(self, item: Range[T]) -> None:
        self._ranges = self._condense_range_list(itertools.chain(self._ranges, [item]))

    def _condense_range_list(self, source_ranges: Iterable[Range[T]]) -> List[Range[T]]:
        """
        Given an iterable of source ranges, return a sorted list of ranges where the ranges have
        been condensed into the smalles amount of disjoint ranges that matches the input.

        For example:

            * [[0,4), [1, 5), [7,9)] -> [[0,5), [7,9)]
        """
        # By sorting the source ranges we can check for overlaps by just looking at whether
        # there is an overlap with the most recent range we added
        ranges: List[Range[T]] = []
        for current_range in sorted(source_ranges):
            if not ranges:
                ranges.append(current_range)
                continue

            previous_range = ranges.pop()
            if (union := previous_range | current_range) is not None:
                ranges.append(union)
            else:
                ranges.extend([previous_range, current_range])

        return ranges

    def discard(self, item: Range[T]) -> None:
        """
        Discarding a range from a range set is equivalent to "cutting away" all the intersections
        of that range with the rangeset.
        """
        ranges: List[Range[T]] = []
        for r in self._ranges:
            difference = r - item
            if difference is None:
                continue
            if isinstance(difference, RangeSet):
                ranges.extend(difference._ranges)
            else:
                ranges.append(difference)

        self._ranges = self._condense_range_list(ranges)

    def pop(self) -> Range[T]:
        """
        Remove and return an arbitrary set element.

        Raises KeyError if the set is empty.
        """
        try:
            elem = self._ranges[0]
        except IndexError:
            raise KeyError

        self.discard(elem)
        return elem

    def __contains__(self, item: Range[T] | T) -> bool:
        """
        Note that this doesn't require exact matching - the target item just has to be inside one
        of the ranges in the set.
        """
        if isinstance(item, Range):
            return self.contains_range(item)
        else:
            return self.contains_item(item)

    def contains_range(self, range: Range[T]) -> bool:
        """Check a Range is fully contained within another Range within this RanegSet."""
        return any((r & range) == range for r in self._ranges)

    def contains_item(self, item: T) -> bool:
        """Check if an item is contained by any Range within this RangeSet."""
        return any(item in range_ for range_ in self)

    def is_disjoint(self, other: "RangeSet[T]") -> bool:
        """
        Check whether this RangeSet is disjoint from the other one.
        """
        return all(a.is_disjoint(b) for a in self._ranges for b in other._ranges)

    def __iter__(self) -> Iterator[Range[T]]:
        yield from self._ranges

    def __len__(self) -> int:
        """
        This isn't particularly useful for consumers, but makes equality work nicely.
        """
        return len(self._ranges)

    def intersection(self, other: Range[T] | RangeSet[T]) -> "RangeSet[T]":
        """
        Return the intersection of this RangeSet and the other Range or RangeSet.
        """
        if isinstance(other, Range):
            other_ranges: Iterable[Range[T]] = [other]
        else:
            other_ranges = other

        return RangeSet(
            intersection
            for r, s in itertools.product(self._ranges, other_ranges)
            if (intersection := r & s) is not None
        )

    def union(self, other: "RangeSet[T]") -> "RangeSet[T]":
        return RangeSet(itertools.chain(self._ranges, other._ranges))

    def __and__(self, other: "RangeSet[T]") -> "RangeSet[T]":
        """
        Wrapper for intersection to allow the & operator to be used.
        """
        return self.intersection(other)

    def __or__(self, other: "RangeSet[T]") -> "RangeSet[T]":
        """
        Wrapper for union to allow the | operator to be used.
        """
        return self.union(other)

    def is_left_finite(self) -> bool:
        return all(r.is_left_finite() for r in self._ranges)

    def is_right_finite(self) -> bool:
        return all(r.is_right_finite() for r in self._ranges)

    def is_finite(self) -> bool:
        return all(r.is_finite() for r in self._ranges)

    def complement(self) -> RangeSet[T]:
        """
        Get a rangeset representing the ranges between the ranges in this rangeset and infinite
        left and right bounds.
        """
        complement: list[Range[T]] = []

        if not self:
            infinite_range: Range[T] = Range(
                None, None, boundaries=RangeBoundaries.EXCLUSIVE_EXCLUSIVE
            )
            return RangeSet([infinite_range])

        if (first_range := self._ranges[0]).is_left_finite():
            complement.append(
                Range(
                    None,
                    first_range.start,
                    boundaries=RangeBoundaries.from_bounds(
                        left_exclusive=True,
                        right_exclusive=(not first_range._is_left_exclusive),
                    ),
                )
            )

        for preceeding_range, current_range in zip(self._ranges[:-1], self._ranges[1:]):
            complement.append(
                Range(
                    preceeding_range.end,
                    current_range.start,
                    boundaries=RangeBoundaries.from_bounds(
                        left_exclusive=(not preceeding_range._is_right_exclusive),
                        right_exclusive=(not current_range._is_left_exclusive),
                    ),
                )
            )

        if (last_range := self._ranges[-1]).is_right_finite():
            complement.append(
                Range(
                    last_range.end,
                    None,
                    boundaries=RangeBoundaries.from_bounds(
                        left_exclusive=(not last_range._is_right_exclusive),
                        right_exclusive=True,
                    ),
                )
            )

        return RangeSet(complement)

    def __neg__(self) -> RangeSet[T]:
        return self.complement()

    def __bool__(self) -> bool:
        return bool(len(self))

    def difference(self, other: RangeSet[T]) -> RangeSet[T]:
        """
        Return a rangeset consisting of the bits of this rangeset that do not intersect the other
        rangeset.
        """
        return self & (-other)

    def __sub__(self, other: RangeSet[T]) -> RangeSet[T]:
        return self.difference(other)


class HalfFiniteRangeSet(RangeSet[T], Generic[T]):
    """
    This subclass is useful when dealing with half-finite intervals as we can offer stronger guarantees
    than we can get with normal RangeSets - mostly around intersections being half-finite etc.
    """

    _ranges: Sequence[HalfFiniteRange[T]]

    def __iter__(self) -> Iterator[HalfFiniteRange[T]]:
        yield from self._ranges

    def intersection(self, other: Range[T] | RangeSet[T]) -> "HalfFiniteRangeSet[T]":
        return cast("HalfFiniteRangeSet[T]", super().intersection(other))

    def pop(self) -> HalfFiniteRange[T]:
        return cast(HalfFiniteRange[T], super().pop())

    def __and__(self, other: RangeSet[T]) -> "HalfFiniteRangeSet[T]":
        return self.intersection(other)


class FiniteRangeSet(RangeSet[T]):
    """
    This subclass is useful when dealing with finite intervals as we can offer stronger guarantees
    than we can get with normal RangeSets - mostly around intersections being finite etc.
    """

    _ranges: Sequence[FiniteRange[T]]

    def __iter__(self) -> Iterator[FiniteRange[T]]:
        yield from self._ranges

    def intersection(self, other: Range[T] | RangeSet[T]) -> "FiniteRangeSet[T]":
        return cast("FiniteRangeSet[T]", super().intersection(other))

    def pop(self) -> FiniteRange[T]:
        """
        Remove and return an arbitrary set element.

        Raises KeyError if the set is empty.
        """
        return cast(FiniteRange[T], super().pop())

    def __and__(self, other: RangeSet[T]) -> "FiniteRangeSet[T]":
        return self.intersection(other)


class FiniteDatetimeRange(FiniteRange[datetime.datetime]):
    """
    This subclass is a helper for a common usecase for ranges - representing finite intervals
    of time.
    """

    __slots__ = ()

    def __init__(self, start: datetime.datetime, end: datetime.datetime):
        """
        Force the boundaries of the range to be [).
        """
        super().__init__(start, end, boundaries=RangeBoundaries.INCLUSIVE_EXCLUSIVE)

    def intersection(
        self, other: Range[datetime.datetime]
    ) -> Optional["FiniteDatetimeRange"]:
        """
        Intersections with finite ranges will always be finite.
        """
        base_intersection = super().intersection(other)
        if base_intersection is None:
            return None

        assert base_intersection.boundaries == RangeBoundaries.INCLUSIVE_EXCLUSIVE
        return FiniteDatetimeRange(base_intersection.start, base_intersection.end)

    def union(self, other: Range[datetime.datetime]) -> Optional["FiniteDatetimeRange"]:
        """
        Unions between two FiniteDatetimeRanges should produce a FiniteDatetimeRange.
        """
        try:
            base_union = super().union(other)
        except ValueError:
            return None

        if base_union is None:
            return None

        assert base_union.boundaries == RangeBoundaries.INCLUSIVE_EXCLUSIVE
        assert base_union.start is not None
        assert base_union.end is not None
        return FiniteDatetimeRange(base_union.start, base_union.end)

    def __and__(
        self, other: Range[datetime.datetime]
    ) -> Optional["FiniteDatetimeRange"]:
        return self.intersection(other)

    @property
    def days(self) -> int:
        """
        Return the number of days between the start and end of the range.
        """
        return (self.end - self.start).days

    @property
    def seconds(self) -> int:
        """
        Return the number of seconds between the start and end of the range.
        """
        return int((self.end - self.start).total_seconds())

    def localize(self, tz: datetime.tzinfo) -> FiniteDatetimeRange:
        """
        Returns the range with boundaries adjusted to the specified timezone.

        See datetime.astimezone for more details.

        Raises:
            ValueError:
                If one or both boundaries are naive (no timezone).
        """
        if not self.start.tzinfo or not self.end.tzinfo:
            raise ValueError("Cannot localize range with naive boundaries")

        return FiniteDatetimeRange(
            self.start.astimezone(tz),
            self.end.astimezone(tz),
        )


class FiniteDateRange(FiniteRange[datetime.date]):
    """
    This subclass is a helper for a common usecase for ranges - representing finite intervals
    of whole days.
    """

    __slots__ = ()

    def __init__(self, start: datetime.date, end: datetime.date):
        """
        Force the boundaries of the range to be [].
        """
        super().__init__(start, end, boundaries=RangeBoundaries.INCLUSIVE_INCLUSIVE)

    def intersection(self, other: Range[datetime.date]) -> Optional["FiniteDateRange"]:
        """
        Intersections with finite ranges will always be finite.
        """
        try:
            base_intersection = super().intersection(other)
        except ValueError:
            # This occurs when calling intersection on an adjacent date range.
            return None
        if base_intersection is None:
            return None

        assert base_intersection.boundaries == RangeBoundaries.INCLUSIVE_INCLUSIVE
        return FiniteDateRange(base_intersection.start, base_intersection.end)

    def union(self, other: Range[datetime.date]) -> Optional["FiniteDateRange"]:
        """
        Unions between two FiniteDateRanges should produce a FiniteDateRange.
        """
        try:
            base_union = super().union(other)
        except ValueError:
            return None

        if base_union is None:
            return None

        assert base_union.boundaries == RangeBoundaries.INCLUSIVE_INCLUSIVE
        assert base_union.start is not None
        assert base_union.end is not None
        return FiniteDateRange(base_union.start, base_union.end)

    def is_disjoint(self, other: Range[datetime.date]) -> bool:
        # Adjacent dates should not be considered disjoint, we extend the other
        # range to allow them to be considered adjacent.
        other_start = other.start
        if other._is_left_inclusive:
            assert other_start is not None
            try:
                other_start -= datetime.timedelta(days=1)
            except OverflowError:
                pass
        other_end = other.end
        if other._is_right_inclusive:
            assert other_end is not None
            try:
                other_end += datetime.timedelta(days=1)
            except OverflowError:
                pass
        return super().is_disjoint(
            Range(start=other_start, end=other_end, boundaries=other.boundaries)
        )

    def __and__(self, other: Range[datetime.date]) -> Optional["FiniteDateRange"]:
        return self.intersection(other)

    @property
    def days(self) -> int:
        """
        Return the number of days between the start and end of the range.
        """
        return (self.end - self.start).days + 1


def get_finite_datetime_ranges_from_timestamps(
    finite_datetime_range: FiniteRange[datetime.datetime],
    timestamps: Iterable[datetime.datetime],
) -> Sequence[FiniteDatetimeRange]:
    """
    Given a datetime range and some timestamps, cut that period into
    multiple points whenever one of the timestamps falls within it.

    Sorts and deduplicates the timestamps first.

    Example:

    - Input:
        - period: ("2021-09-01 00:00:00", "2021-10-01 00:00:00")
        - timestamps: ["2021-09-10 00:00:00", "2021-09-16 00:00:00", "2021-09-23 00:00:00"]
    - Return:
        [
            ("2021-09-01 00:00:00", "2021-09-10 00:00:00"),
            ("2021-09-10 00:00:00", "2021-09-16 00:00:00"),
            ("2021-09-16 00:00:00", "2021-09-23 00:00:00"),
            ("2021-09-23 00:00:00", "2021-10-01 00:00:00"),
        ]
    """
    timestamps_in_range = sorted(
        {timestamp for timestamp in timestamps if timestamp in finite_datetime_range}
    )

    return [
        FiniteDatetimeRange(start, end)
        for start, end in zip(
            [finite_datetime_range.start, *timestamps_in_range],
            [*timestamps_in_range, finite_datetime_range.end],
        )
        # We don't need to split when a timestamp is exactly at the start
        # or end of the period.
        if start != end
    ]


def any_overlapping(ranges: Iterable[Range[T]]) -> bool:
    """Return true if any of the passed Ranges are overlapping."""
    ranges = list(ranges)
    if not ranges:
        return False
    range_set = RangeSet[T]([ranges[0]])
    for range in ranges[1:]:
        if range_set.intersection(range):
            return True
        range_set.add(range)
    return False


def as_finite_datetime_periods(
    periods: Iterable[HalfFiniteDatetimeRange | DatetimeRange],
) -> Sequence[FiniteDatetimeRange]:
    """
    Casts the given date/time periods as finite periods.

    This is useful when working with potentially infinite ranges that are
    known to be finite e.g. due to intersection with a finite range.

    Raises:
        ValueError: If one or more periods is not finite.
    """
    finite_periods = []

    for period in periods:
        if period.start is None or period.end is None:
            raise ValueError("Period is not finite at start or end or both")

        finite_periods += [FiniteDatetimeRange(period.start, period.end)]

    return finite_periods


def iterate_over_months(
    period: FiniteDatetimeRange, *, tz: datetime.tzinfo
) -> Iterator[FiniteDatetimeRange]:
    """
    Generate a sequence of finite datetime ranges spanning months during the period.

    Ranges span each whole month between the start and end times, inclusive.

    ie: given a period of (15/01, 15/03) -> [
        (15/01->01/02),
        (01/02->01/03),
        (01/03->15/03)
    ]
    """
    start_at = localtime.as_localtime(period.start, tz=tz)
    end_at = localtime.as_localtime(period.end, tz=tz)

    while True:
        next_start = start_at + relativedelta.relativedelta(months=1, day=1)
        this_end = next_start

        if end_at <= this_end:
            yield FiniteDatetimeRange(start_at, end_at)
            return

        yield FiniteDatetimeRange(start_at, this_end)
        start_at = next_start


def date_range_for_midnight_range(
    range: FiniteDatetimeRange,
) -> FiniteDateRange:
    """
    Returns the date range of a midnight-aligned datetime range.

    This can be useful where a range is available at datetime granularity,
    but is used in functions that operate at date granularity.

    Raises:
        ValueError:
            If the range boundaries are in different timezeones.
            If the range boundaries are not midnight-aligned.
    """
    # First check range timezone is uniform.
    if range.start.tzinfo != range.end.tzinfo:
        raise ValueError("Start and end in different timezones")

    # Check datetimes are both midnight-aligned.
    if range.start.time() != datetime.time(0, 0):
        raise ValueError("Start of range is not midnight-aligned")

    if range.end.time() != datetime.time(0, 0):
        raise ValueError("End of range is not midnight-aligned")

    return FiniteDateRange(
        range.start.date(),
        range.end.date() - datetime.timedelta(days=1),
    )



"""
This file contains helper functions necessary for using the s3 select functionality.

Note:
    - Many comments have been taken directly from the boto3 [docs]
(https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.select_object_content)
"""

from __future__ import annotations

import dataclasses
import enum
from typing import Any


class S3SelectException(Exception):
    pass


class FileHeaderInfo(str, enum.Enum):
    """
    A bucket of strings used by the `select_object_content` function to read a CSV file.
    """

    # First line is not a header.
    NONE = "NONE"

    # First line is a header, but you can't use the header values to indicate the column in an expression.
    # You can use column position (such as _1, _2, …) to indicate the column (SELECT s._1 FROM OBJECT s ).
    IGNORE = "IGNORE"

    # First line is a header, and you can use the header value to identify a column in an expression
    # (SELECT "name" FROM OBJECT ).
    USE = "USE"


class CompressionType(str, enum.Enum):
    # Specifies object's compression format.
    NONE = "NONE"
    GZIP = "GZIP"
    BZIP2 = "BZIP2"


class QuoteFields(str, enum.Enum):
    # Always use quotation marks for output fields.
    ALWAYS = "ALWAYS"
    # Use quotation marks for output fields when needed.
    ASNEEDED = "ASNEEDED"


@dataclasses.dataclass(frozen=True)
class BaseSerializer:
    def to_dict(self) -> dict[str, Any]:
        """
        A helper method to remove None values from the dict and ensures all enum fields use their values
        instead of an enum instance.
        """

        temp_dict = {k: v for k, v in dataclasses.asdict(self).items() if v}

        for k, v in temp_dict.items():
            if hasattr(v, "value"):
                temp_dict[k] = v.value
        return temp_dict


@dataclasses.dataclass(frozen=True)
class CSVInputSerializer(BaseSerializer):
    # Describes the first line of input
    FileHeaderInfo: FileHeaderInfo
    # A single character used to indicate that a row should be ignored when the character is present at the start of
    # that row. You can specify any character to indicate a comment line.
    Comments: str | None = None
    # A single character used for escaping the quotation mark character inside an already escaped value.
    # For example, the value """ a , b """ is parsed as " a , b " .
    QuoteEscapeCharacter: str | None = None
    # A single character used to separate individual records in the input. Instead of the default value,
    # you can specify an arbitrary delimiter.
    RecordDelimiter: str | None = "\n"
    # A single character used to separate individual fields in a record. You can specify an arbitrary delimiter.
    FieldDelimiter: str | None = ","
    # A single character used for escaping when the field delimiter is part of the value. For example, if the
    # value is a, b , Amazon S3 wraps this field value in quotation marks, as follows: " a , b " .
    QuoteCharacter: str | None = None
    # Specifies that CSV field values may contain quoted record delimiters and such records should be allowed.
    # Default value is FALSE. Setting this value to TRUE may lower performance.
    AllowQuotedRecordDelimiter: bool | None = False


@dataclasses.dataclass(frozen=True)
class ParquetInputSerializer(BaseSerializer):
    pass


@dataclasses.dataclass(frozen=True)
class CSVOutputSerializer(BaseSerializer):
    # Indicates whether to use quotation marks around output fields.
    QuoteFields: QuoteFields | None = None
    # The value used to separate individual fields in a record. You can specify an arbitrary delimiter.
    FieldDelimiter: str | None = ","
    # A single character used for escaping when the field delimiter is part of the value. For example,
    # if the value is a, b, Amazon S3 wraps this field value in quotation marks, as follows: " a , b " .
    QuoteCharacter: str | None = None
    # The single character used for escaping the quote character inside an already escaped value.
    QuoteEscapeCharacter: str | None = None
    # A single character used to separate individual records in the output. Instead of the default value,
    # you can specify an arbitrary delimiter.
    RecordDelimiter: str | None = "\n"


@dataclasses.dataclass(frozen=True)
class JSONOutputSerializer(BaseSerializer):
    # A single character used to separate individual records in the output. Instead of the default value,
    # you can specify an arbitrary delimiter.
    RecordDelimiter: str | None = "\n"


@dataclasses.dataclass(frozen=True)
class ScanRange(BaseSerializer):
    """
    Specifies the byte range of the object to get the records from.

    - Provide a start and an end value to perform an inclusive search.
    - Provide only the start value to process search after the given X records.
    - Provide only the end value to search the last X records of the file.
    """

    Start: int | None = None
    End: int | None = None


def get_serializers_for_csv_file(
    *,
    input_serializer: CSVInputSerializer,
    compression_type: CompressionType,
    output_serializer: CSVOutputSerializer | JSONOutputSerializer,
    scan_range: ScanRange | None = None,
) -> dict[str, dict[str, Any]]:
    """
    Returns input and output serialization dictionaries that should be used to perform a select_object_content query.

    Note: It is recommended to use JSONOutputSerializer over CSVOutputSerializer while using ScanRange queries
    because the former prevents premature chunked data. In other words, because scan range queries loop over the file
    using byte range, you can expect the JSON output format to at least return one single row in each iteration. With
    CSV output format, there is a high chance of receiving output that has been chunked to a certain byte.
    """
    output_format = output_serializer.to_dict()

    if isinstance(output_serializer, CSVOutputSerializer):
        output_serialization_format = {"CSV": output_format}
    else:
        output_serialization_format = {"JSON": output_format}

    temp_dict: dict[str, dict[str, Any]] = {
        "input_serialization": {
            "CSV": input_serializer.to_dict(),
            "CompressionType": compression_type.value,
        },
        "output_serialization": output_serialization_format,
    }

    if scan_range:
        if compression_type == CompressionType.GZIP:
            raise S3SelectException(
                "Scan range queries are not supported on objects with type GZIP."
            )
        temp_dict["ScanRange"] = scan_range.to_dict()

    return temp_dict


def get_serializers_for_parquet_file(
    output_serializer: JSONOutputSerializer | CSVOutputSerializer,
) -> dict[str, dict[str, Any]]:
    """
    Returns input and output serialization dictionaries that should be used to perform a select_object_content query.

    Parquet files do not support input serialization or scan ranges.
    """
    return {
        "input_serialization": {
            "Parquet": {},
        },
        "output_serialization": {
            "JSON"
            if isinstance(output_serializer, JSONOutputSerializer)
            else "CSV": output_serializer.to_dict()
        },
    }



from xocto import pact_testing


def test_pact_service():
    options = pact_testing.PactOptions(
        broker_url="https://pact-broker.example.com",
        broker_username="username",
        broker_password="password",
        consumer_name="consumer_name",
        consumer_version="1.0.0",
        provider_name="provider_name",
        log_path="/tmp/pact/",
    )
    pact_service = pact_testing.pact_service(
        options=options,
        publish_to_broker=True,
    )
    assert pact_service.broker_base_url == "https://pact-broker.example.com"
    assert pact_service.broker_username == "username"
    assert pact_service.broker_password == "password"
    assert pact_service.provider.name == "provider_name"
    assert pact_service.consumer.name == "consumer_name"
    assert pact_service.consumer.version == "1.0.0"
    assert pact_service.publish_to_broker is True
    assert pact_service.pact_dir == "/tmp/pact/"
    assert pact_service.log_dir == "/tmp/pact/"



"""
Adapter fields for postgres range types that return xocto.Range python objects.
"""

import datetime
import json
from typing import Any, Optional, Union

from django.contrib.postgres import fields as pg_fields
from django.contrib.postgres.fields import ranges as pg_ranges
from django.contrib.postgres.fields import utils as pg_utils
from django.db import models

from xocto import localtime, ranges


class FiniteDateRangeField(pg_fields.DateRangeField):
    """
    A DateRangeField with Inclusive-Inclusive [] bounds.

    The underlying postgres type will always store the range as Inclusive-Exclusive [).
    This field will always translate to [].

    Accepts and returns xocto.ranges.FiniteDateRange objects.
    """

    def get_prep_value(
        self, value: Any
    ) -> Optional[Union[pg_ranges.DateRange, datetime.date]]:
        if value is None:
            return None
        if isinstance(value, ranges.FiniteDateRange):
            return pg_ranges.DateRange(lower=value.start, upper=value.end, bounds="[]")
        if (
            isinstance(value, datetime.date)
            # Don't allow datetime (datetime is a subclass of date!)
            and not isinstance(value, datetime.datetime)
        ):
            return value
        raise TypeError(
            "FiniteDateRangeField may only accept FiniteDateRange or date objects."
        )

    def from_db_value(
        self,
        value: Optional[pg_ranges.DateRange],
        expression: object,
        connection: object,
    ) -> Optional[ranges.FiniteDateRange]:
        if value is None:
            return None
        return ranges.FiniteDateRange(
            start=value.lower, end=self._upper_to_inclusive(value)
        )

    def to_python(self, value: Optional[str]) -> Optional[ranges.FiniteDateRange]:
        if value is None:
            return None
        obj = json.loads(value)
        return ranges.FiniteDateRange(
            start=self.base_field.to_python(obj["start"]),
            end=self.base_field.to_python(obj["end"]),
        )

    def value_to_string(self, obj: models.Model) -> Optional[str]:
        value: Optional[ranges.FiniteDateRange] = self.value_from_object(obj)
        if value is None:
            return None
        if not isinstance(value, ranges.FiniteDateRange):
            raise TypeError(
                "FiniteDateRangeField may only accept FiniteDateRange objects."
            )
        base_field = self.base_field
        start = pg_utils.AttributeSetter(base_field.attname, value.start)
        end = pg_utils.AttributeSetter(base_field.attname, value.end)
        return json.dumps(
            {
                "start": base_field.value_to_string(start),
                "end": base_field.value_to_string(end),
            }
        )

    def _upper_to_inclusive(self, value: pg_ranges.DateRange) -> datetime.date:
        """
        Transform an exclusive upper bound to an inclusive bound.

        DateRanges (as all discrete ranges) are always stored as [) in the database, even if the
        input was []. [X, Y] is equivalent to [X, Y+1) in the database. When we read the value
        from the database, we need to subtract 1 day from the upper bound to return it to inclusive.
        """
        if value.upper_inc:
            return value.upper
        return value.upper - datetime.timedelta(days=1)


class _LocaliserMixin:
    def __init__(
        self, *args: Any, timezone: Optional[datetime.tzinfo] = None, **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        self._timezone = timezone

    def localise(self, value: datetime.datetime) -> datetime.datetime:
        return localtime.as_localtime(value, self._timezone)


class FiniteDateTimeRangeField(_LocaliserMixin, pg_fields.DateTimeRangeField):
    """
    A DateTimeRangeField with Inclusive-Exclusive [) bounds that aren't infinite.

    Accepts and returns xocto.ranges.FiniteDatetimeRange objects. Values are always
    timezone-aware, and will be converted to the timezone specified in django settings.
    """

    def get_prep_value(
        self, value: Any
    ) -> Optional[Union[pg_ranges.DateTimeTZRange, datetime.datetime]]:
        if value is None:
            return None
        if isinstance(value, ranges.FiniteDatetimeRange):
            return pg_ranges.DateTimeTZRange(
                lower=value.start, upper=value.end, bounds="[)"
            )
        if isinstance(value, datetime.datetime):
            return value
        raise TypeError(
            "FiniteDateTimeRangeField may only accept FiniteDatetimeRange or datetime objects."
        )

    def from_db_value(
        self,
        value: Optional[pg_ranges.DateTimeTZRange],
        expression: object,
        connection: object,
    ) -> Optional[ranges.FiniteDatetimeRange]:
        if value is None:
            return None
        return ranges.FiniteDatetimeRange(
            start=self.localise(value.lower),
            end=self.localise(value.upper),
        )

    def to_python(self, value: Optional[str]) -> Optional[ranges.FiniteDatetimeRange]:
        if value is None:
            return None
        obj = json.loads(value)
        return ranges.FiniteDatetimeRange(
            start=self.localise(self.base_field.to_python(obj["start"])),
            end=self.localise(self.base_field.to_python(obj["end"])),
        )

    def value_to_string(self, obj: models.Model) -> Optional[str]:
        value: Optional[ranges.FiniteDatetimeRange] = self.value_from_object(obj)
        if value is None:
            return None
        if not isinstance(value, ranges.FiniteDatetimeRange):
            raise TypeError(
                "FiniteDateTimeRangeField may only accept FiniteDatetimeRange objects."
            )
        base_field = self.base_field
        start = pg_utils.AttributeSetter(base_field.attname, value.start)
        end = pg_utils.AttributeSetter(base_field.attname, value.end)
        return json.dumps(
            {
                "start": base_field.value_to_string(start),
                "end": base_field.value_to_string(end),
            }
        )


class HalfFiniteDateTimeRangeField(_LocaliserMixin, pg_fields.DateTimeRangeField):
    """
    A DateTimeRangeField with Inclusive-Exclusive [) bounds that allows an infinite/open upper bound.

    Accepts and returns xocto.ranges.HalfFiniteDatetimeRange objects. Values are always
    timezone-aware, and will be converted to the timezone specified in django settings.
    """

    def get_prep_value(
        self, value: Any
    ) -> Optional[Union[pg_ranges.DateTimeTZRange, datetime.datetime]]:
        if value is None:
            return None
        if self._is_half_finite_datetime_range(value):
            return pg_ranges.DateTimeTZRange(
                lower=value.start, upper=value.end, bounds="[)"
            )
        if isinstance(value, datetime.datetime):
            return value
        raise TypeError(
            "HalfFiniteDateTimeRangeField may only accept HalfFiniteDateTimeRangeField or datetime objects."
        )

    def from_db_value(
        self,
        value: Optional[pg_ranges.DateTimeTZRange],
        expression: object,
        connection: object,
    ) -> Optional[ranges.HalfFiniteDatetimeRange]:
        if value is None:
            return None
        return ranges.HalfFiniteDatetimeRange(
            start=self.localise(value.lower),
            end=self.localise(value.upper) if value.upper else None,
        )

    def to_python(
        self, value: Optional[str]
    ) -> Optional[ranges.HalfFiniteDatetimeRange]:
        if value is None:
            return None
        obj = json.loads(value)
        end = self.base_field.to_python(obj["end"])
        return ranges.HalfFiniteDatetimeRange(
            start=self.localise(self.base_field.to_python(obj["start"])),
            end=self.localise(end) if end else None,
        )

    def value_to_string(self, obj: models.Model) -> Optional[str]:
        value: Optional[ranges.HalfFiniteDatetimeRange] = self.value_from_object(obj)
        if value is None:
            return None
        if not isinstance(value, ranges.HalfFiniteRange):
            raise TypeError(
                "HalfFiniteDateTimeRangeField may only accept HalfFiniteDatetimeRange objects."
            )
        base_field = self.base_field
        start = pg_utils.AttributeSetter(base_field.attname, value.start)
        end = pg_utils.AttributeSetter(base_field.attname, value.end)
        return json.dumps(
            {
                "start": base_field.value_to_string(start),
                "end": base_field.value_to_string(end),
            }
        )

    def _is_half_finite_datetime_range(self, value: Any) -> bool:
        # HalfFiniteDatetimeRange is a subscripted generic and may not be checked with
        # isinstance directly. So, we check for it's parent class and attribute values.
        if not isinstance(value, ranges.HalfFiniteRange):
            return False
        if not isinstance(value.start, datetime.datetime):
            return False
        if value.end is not None and not isinstance(value.end, datetime.datetime):
            return False
        return True



from __future__ import annotations

import decimal
import random
from typing import TypeVar

from xocto.types import generic


def quantise(
    number: int | float | str, base: int, rounding: str = decimal.ROUND_HALF_EVEN
) -> int:
    """
    Round a number to an arbitrary integer base. For example:
    >>> quantise(256, 5)
    255

    Note that numbers equal to half of the rounding amount will always
    round down. So:
    >>> quantise(15, 30)
    0
    """
    assert base > 0
    rounded_result = base * (decimal.Decimal(number) / base).quantize(
        decimal.Decimal("1."), rounding=rounding
    )
    return int(rounded_result)


def truncate_decimal_places(value: decimal.Decimal, places: int = 1) -> float:
    """
    Truncate a float (i.e round towards zero) to a given number of decimal places.

    NB: Takes a decimal but returns a float!

    >>> truncate_decimal_places(12.364, 1)
    12.3

    >>> round_decimal_places(-12.364, 1)
    -12.3 # -12.3 is bigger than -12.4

    >>> round_decimal_places(12.364, 0)
    12.0 # rounding to 0 returns float with no decmial part
    """

    if places == 0:
        quantize_string = "1"
    else:
        quantize_string = "0." + ((places - 1) * "0") + "1"

    exponent = decimal.Decimal(quantize_string)
    decimal_result = value.quantize(exponent, rounding=decimal.ROUND_DOWN)
    return float(decimal_result)


def round_decimal_places(
    value: decimal.Decimal, places: int = 1, rounding: str = decimal.ROUND_HALF_UP
) -> decimal.Decimal:
    """
    Round a decimal to a given number of decimal places using a given rounding method.

    By default, we use half-up rounding so that parts from half way between the given rounding
    precision will be rounded up towards the greater number.

    This differs from the default rounding since Python 3.0 which is also used elsewhere in Kraken
    (which use "banker's"/half-even rounding, which is considered by IEEE 754 to be the recommended
    default for decimal).

    >>> round_decimal_places(12.35, 1)
    12.4

    >>> round_decimal_places(-12.35, 1)
    -12.3 #-12.3 is bigger than -12.4
    """

    if places == 0:
        quantize_string = "1"
    else:
        quantize_string = "0." + ((places - 1) * "0") + "1"

    return value.quantize(decimal.Decimal(quantize_string), rounding=rounding)


def round_to_integer(
    value: decimal.Decimal, rounding: str = decimal.ROUND_HALF_UP
) -> int:
    """
    Round a decimal to the nearest integer, using a given rounding method.

    By default, we use half-up rounding.

    This differs from the default rounding since Python 3.0 which is also used elsewhere in Kraken
    (which use "banker's"/half-even rounding, which is considered by IEEE 754 to be the recommended
    default for decimal).
    """
    return int(round_decimal_places(value, places=0, rounding=rounding))


T = TypeVar("T")


def clip_to_range(
    val: generic.Comparable[T],
    *,
    minval: generic.Comparable[T],
    maxval: generic.Comparable[T],
) -> generic.Comparable[T]:
    """
    Clip the value to the min and max values given.

    Values to be compared must be of the same type.

    Example usage:
        >>> clip_to_range(10, minval=20, maxval=25)
        20
        >>> clip_to_range(date(2020, 1, 4), minval=date(2020, 1, 1), maxval=date(2020, 1, 3))
        date(2020, 1, 3)
        >>> clip_to_range(1.5, minval=1.0, maxval=2.0)
        1.5
    """
    assert not minval > maxval, "minval must not be greater than maxval"
    if val < minval:
        return minval
    elif val > maxval:
        return maxval
    else:
        return val


def remove_exponent(d: decimal.Decimal) -> decimal.Decimal:
    """
    Util function for removing the exponent and trailing zeroes of a decimal.

    From https://docs.python.org/3/library/decimal.html#decimal-faq
    """
    return d.quantize(decimal.Decimal(1)) if d == d.to_integral() else d.normalize()


def random_int(length: int) -> int:
    """
    Return a pseudo-random integer based on the provided `length`.

        >>> random_int(3)
        114
    """
    if length < 2:
        raise ValueError("length must be greater than or equal to 2")
    range_start = 10 ** (length - 1)
    range_end = (10**length) - 1
    return random.randint(range_start, range_end)



import time
from typing import Any


__all__ = ["Timer"]


class Timer(object):
    """
    Context manager to allow easy timing of events.
    """

    def __enter__(self) -> "Timer":
        self.start = time.time()
        return self

    def __exit__(self, *args: Any) -> None:
        self.end = time.time()
        self.duration_in_ms = (self.end - self.start) * 1000



import datetime
import decimal
import zoneinfo

import pytest
import time_machine
from dateutil import relativedelta
from django.conf import settings
from django.test import override_settings
from django.utils import timezone

from tests import factories
from xocto import localtime


class TestNow:
    def test_now_is_in_correct_timezone(self):
        now = localtime.now()
        assert str(now.tzinfo) == settings.TIME_ZONE


class TestSecondsInTheFuture:
    def test_seconds_in_future(self):
        with time_machine.travel("2020-01-01 12:00:00.000", tick=False):
            assert localtime.seconds_in_the_future(1) == localtime.as_localtime(
                localtime.datetime_.datetime(2020, 1, 1, 12, 0, 1, tzinfo=localtime.UTC)
            )
            assert localtime.seconds_in_the_future(1.5) == localtime.as_localtime(
                localtime.datetime_.datetime(
                    2020, 1, 1, 12, 0, 1, 500000, tzinfo=localtime.UTC
                )
            )


class TestSecondsInThePast:
    def test_seconds_in_past(self):
        with time_machine.travel("2020-01-01 12:00:00.000", tick=False):
            assert localtime.seconds_in_the_past(1) == localtime.as_localtime(
                localtime.datetime_.datetime(
                    2020, 1, 1, 11, 59, 59, tzinfo=localtime.UTC
                )
            )
            assert localtime.seconds_in_the_past(1.5) == localtime.as_localtime(
                localtime.datetime_.datetime(
                    2020, 1, 1, 11, 59, 58, 500000, tzinfo=localtime.UTC
                )
            )


class TestDate:
    def test_date_calculation_near_midnight_during_bst(self):
        near_midnight_in_utc = datetime.datetime(
            2016, 6, 1, 23, 50, 0, tzinfo=localtime.UTC
        )
        assert localtime.date(near_midnight_in_utc) == (
            near_midnight_in_utc.date() + datetime.timedelta(days=1)
        )

    def test_date_calculation_near_midnight_outside_of_bst(self):
        near_midnight_in_utc = datetime.datetime(
            2016, 11, 1, 23, 50, 0, tzinfo=localtime.UTC
        )
        assert localtime.date(near_midnight_in_utc) == near_midnight_in_utc.date()

    @pytest.mark.parametrize("tz", (zoneinfo.ZoneInfo("Etc/GMT-10"), localtime.UTC))
    def test_date_calculation_specifying_other_timezone(self, tz):
        near_midnight = datetime.datetime(2016, 6, 1, 23, 50, 0, tzinfo=tz)
        assert localtime.date(near_midnight, tz=tz) == (near_midnight.date())

    def test_datetime_not_supplied(self):
        """
        Check that we do not fallback to today if a datetime is not passed to the function -
        we have localtime.today for that.
        """
        with pytest.raises(
            TypeError, match="You must supply a datetime to localtime.date"
        ):
            localtime.date(None)


class TestDateOfDayBefore:
    @override_settings(TIME_ZONE="Europe/London")
    @pytest.mark.parametrize(
        "dt, date_of_day_before",
        [
            # Local time midnight: does what you'd expect
            (factories.local.dt("2019-06-01 00:00:00"), factories.date("2019-05-31")),
            # Note that this next example is *the same time*, but expressed in UTC. So again, it
            # correctly returns the same date. But you might not be expecting it when you see it.
            (factories.utc.dt("2019-05-31 23:00:00"), factories.date("2019-05-31")),
            # A couple of 'normal' examples
            (factories.local.dt("2019-01-23 05:30:00"), factories.date("2019-01-22")),
            (factories.utc.dt("2019-01-23 05:30:00"), factories.date("2019-01-22")),
        ],
    )
    def test_returns_expected_results(self, dt, date_of_day_before):
        assert localtime.date_of_day_before(dt) == date_of_day_before


class TestMidnight:
    def test_midnight_calculation_for_bst_date(self):
        date = datetime.date(2016, 7, 2)
        midnight = localtime.midnight(date)
        assert midnight.tzinfo is not None
        assert midnight.hour == 0

        midnight_utc = midnight.astimezone(localtime.UTC)
        assert midnight_utc.hour == 23

    def test_midnight_calculation_for_utc_dt(self):
        dt = factories.utc.dt("2017-12-20T16:00")

        assert localtime.midnight(dt) == localtime.datetime(2017, 12, 20, 0, 0)

    def test_midnight_calculation_without_date_uses_today(self):
        midnight = localtime.midnight()
        assert midnight.date() == localtime.today()
        assert midnight.tzinfo is not None
        assert midnight.hour == 0

    def test_convert_date_to_midnight_and_back(self):
        date = datetime.date(2016, 7, 2)
        midnight = localtime.midnight(date)
        midnight_utc = midnight.astimezone(localtime.UTC)
        assert localtime.date(midnight_utc) == date

    def test_midnight_in_different_timezone(self):
        aus_time = zoneinfo.ZoneInfo("Etc/GMT-10")

        with time_machine.travel(
            datetime.datetime(2020, 2, 2, 1, tzinfo=aus_time), tick=False
        ):
            result = localtime.midnight(tz=aus_time)

        assert result == datetime.datetime(2020, 2, 2, 0, 0, tzinfo=aus_time)

    def test_doesnt_change_date_of_already_midnight_datetime(self):
        """
        Check that if we pass a datetime to `midnight()` which is *already* midnight in the
        relevant timezone, it doesn't get changed.
        """
        midnight = datetime.datetime(2020, 6, 1, 23, tzinfo=datetime.timezone.utc)

        # We'll assert the same thing three ways for clarity:
        assert (
            localtime.midnight(midnight).date()
            == localtime.as_localtime(midnight).date()
        )
        assert localtime.midnight(midnight).date() == datetime.date(2020, 6, 2)
        assert localtime.midnight(midnight) == midnight

    @override_settings(
        TIME_ZONE="Australia/Sydney"
    )  # set the django default/current timezone
    @pytest.mark.parametrize(
        "naive_datetime,expected_midnight",
        [
            (
                datetime.datetime(2021, 6, 17, 18, 0, 0),
                datetime.datetime(2021, 6, 17, 0, 0, 0),
            ),
            (
                datetime.datetime(2021, 6, 17, 23, 30, 0),
                datetime.datetime(2021, 6, 17, 0, 0, 0),
            ),
            (
                datetime.datetime(2021, 6, 18, 0, 0, 0),
                datetime.datetime(2021, 6, 18, 0, 0, 0),
            ),
            (
                datetime.datetime(2021, 6, 18, 0, 30, 0),
                datetime.datetime(2021, 6, 18, 0, 0, 0),
            ),
            (
                datetime.datetime(2021, 6, 18, 6, 0, 0),
                datetime.datetime(2021, 6, 18, 0, 0, 0),
            ),
        ],
    )
    def test_localtime_midnight_calculation_for_naive_datetime_and_no_timezone(
        self, naive_datetime, expected_midnight
    ):
        """
        Test the behavior of localtime.midnight() when passed a naive datetime and NOT passed a tz

        We prefer to not pass a naive datetime to midnight(), but if you do,
        this is the behaviour with no tz argument:
        1. the naive datetime is assumed to be in the current Django timezone
        2. get the date of that datetime
        3. return midnight at the start of that date, in the current Django timezone.
        """
        # attach the current Sydney timezone to the expected midnight
        expected_midnight = timezone.make_aware(expected_midnight)

        actual_midnight = localtime.midnight(naive_datetime)

        assert actual_midnight == expected_midnight
        assert str(actual_midnight.tzinfo) == "Australia/Sydney"

    @override_settings(
        TIME_ZONE="Australia/Sydney"
    )  # set the django default/current timezone
    @pytest.mark.parametrize(
        "naive_datetime,specified_timezone,expected_midnight",
        [
            (
                # this is assumed to be in the current django timezone (Sydney)
                datetime.datetime(2021, 6, 17, 18, 0, 0),
                # then converted to UTC - moved 10 hours back - still same date
                "UTC",
                # result is midnight at start of the date
                datetime.datetime(2021, 6, 17, 0, 0, 0),
            ),
            (
                # this is assumed to be in the current django timezone (Sydney)
                datetime.datetime(2021, 6, 21, 5, 0, 0),
                # then converted to UTC - moved 10 hours back - changes to previous date
                "UTC",
                # midnight returned is previous date - this result can be unexpected
                datetime.datetime(2021, 6, 20, 0, 0, 0),
            ),
            (
                # this is assumed to be in the current django timezone (Sydney)
                datetime.datetime(2021, 6, 24, 23, 0, 0),
                # then converted to Pacific/Auckland - moved 2 hours ahead - changes to next date
                "Pacific/Auckland",
                # midnight returned is next date - this result can be unexpected
                datetime.datetime(2021, 6, 25, 0, 0, 0),
            ),
            (
                # This example was the cause of a bug in Australia:
                # this is assumed to be in the current django timezone (Sydney),
                # which on this date is UTC+11 (DST)
                datetime.datetime(2021, 2, 16, 0, 0, 0),
                # then converted to UTC+10 (the INDUSTRY_TIMEZONE in Australia)
                # which moves it 1 hour back - changes to previous date
                "Etc/GMT-10",
                # midnight returned is previous date - this result can be unexpected
                datetime.datetime(2021, 2, 15, 0, 0, 0),
            ),
            (
                # this is assumed to be in the current django timezone (Sydney),
                # which on this date is UTC+10
                datetime.datetime(2021, 7, 16, 0, 0, 0),
                # then converted to UTC+10 (the INDUSTRY_TIMEZONE in Australia)
                # no change - date is kept the same
                "Etc/GMT-10",
                # midnight returned is same date
                datetime.datetime(2021, 7, 16, 0, 0, 0),
            ),
        ],
    )
    def test_localtime_midnight_calculation_for_naive_datetime_and_specified_timezone(
        self, naive_datetime, specified_timezone, expected_midnight
    ):
        """
        Test the behavior of localtime.midnight() when passed a naive datetime and a timezone

        We prefer to not pass a naive datetime to midnight(), but if you do,
        this is the behaviour when tz argument is specified:
        1. the naive datetime is assumed to be in the current Django timezone
            (NOT the passed timezone)
        2. convert that moment in time to the specified timezone
        3. get the date of that datetime
        3. return midnight at the start of that date, in the specified timezone.
        """
        specified_timezone_obj = zoneinfo.ZoneInfo(specified_timezone)
        # attach the specified timezone to the expected midnight
        expected_midnight = timezone.make_aware(
            expected_midnight, timezone=specified_timezone_obj
        )

        actual_midnight = localtime.midnight(naive_datetime, tz=specified_timezone_obj)

        assert actual_midnight == expected_midnight
        assert str(actual_midnight.tzinfo) == specified_timezone


class TestMidday:
    def test_midday_calculation(self):
        date = datetime.date(2016, 7, 2)
        midday = localtime.midday(date)
        assert midday.tzinfo is not None
        assert midday.hour == 12
        assert midday.minute == 0

    def test_midday_calculation_without_date_uses_today(self):
        midday = localtime.midday()
        assert midday.date() == localtime.today()
        assert midday.tzinfo is not None
        assert midday.hour == 12
        assert midday.minute == 0

    def test_midday_in_different_timezone(self):
        aus_time = zoneinfo.ZoneInfo("Etc/GMT-10")

        with time_machine.travel(
            datetime.datetime(2020, 2, 2, 1, tzinfo=aus_time), tick=False
        ):
            result = localtime.midday(tz=aus_time)

        assert result == datetime.datetime(2020, 2, 2, 12, 0, tzinfo=aus_time)


class TestLatest:
    def test_latest_calculation(self):
        date = datetime.date(2016, 7, 2)
        latest = localtime.latest(date)
        assert latest.tzinfo is not None
        assert latest.hour == 23
        assert latest.minute == 59
        assert latest.second == 59
        assert latest.microsecond == 999999

    def test_latest_calculation_without_date_uses_today(self):
        latest = localtime.latest()
        assert latest.date() == localtime.today()
        assert latest.tzinfo is not None
        assert latest.hour == 23
        assert latest.minute == 59
        assert latest.second == 59
        assert latest.microsecond == 999999


class TestDateTime:
    def test_datetime_creation(self):
        dt = localtime.datetime(2016, 8, 5)
        assert dt.hour == 0
        utc_dt = dt.astimezone(localtime.UTC)
        assert utc_dt.hour == 23

    def test_dst_ambiguity(self):
        # 2020-10-25T01:30 is an ambiguous dt in Europe/London as its in the period when clocks go
        # back (so it occurs twice).
        dt = localtime.datetime(2020, 10, 25, 1, 30)
        assert dt.hour == 1
        assert dt.minute == 30
        assert dt.dst() == datetime.timedelta(seconds=3600)
        utc_dt = dt.astimezone(localtime.UTC)
        assert utc_dt.hour == 0


class TestAsLocaltime:
    def test_conversion_of_gmt_dt(self):
        non_dst_datetime = datetime.datetime(2016, 1, 1, 23, 1, tzinfo=localtime.UTC)
        converted_non_dst_datetime = localtime.as_localtime(non_dst_datetime)

        # The two dates are 'equal' and their day and hour attributes differ
        assert converted_non_dst_datetime == non_dst_datetime
        assert converted_non_dst_datetime.day == non_dst_datetime.day
        assert converted_non_dst_datetime.hour == non_dst_datetime.hour

    def test_conversion_of_bst_dt(self):
        dst_datetime = datetime.datetime(2016, 5, 1, 23, 1, tzinfo=localtime.UTC)
        converted_dst_datetime = localtime.as_localtime(dst_datetime)

        # The two dates are 'equal' but their day and hour attributes differ
        assert converted_dst_datetime == dst_datetime
        assert converted_dst_datetime.day == dst_datetime.day + 1
        assert converted_dst_datetime.hour == dst_datetime.hour - 23

    def test_converts_utc_to_europe_london(self):
        utc_dt = factories.utc.dt("2018-05-28T23:30")
        assert str(utc_dt.tzinfo) == "UTC"

        local_dt = localtime.as_localtime(utc_dt)
        assert str(local_dt.tzinfo) == "Europe/London"


class TestIsUTC:
    @pytest.mark.parametrize("tzinfo", [datetime.timezone.utc, localtime.UTC])
    def test_is_utc(self, tzinfo):
        now = datetime.datetime(2020, 1, 1, tzinfo=tzinfo)
        assert localtime.is_utc(now)

    def test_is_not_utc(self):
        now = datetime.datetime(2020, 1, 1, tzinfo=localtime.LONDON)
        assert not localtime.is_utc(now)


class TestIsLocalime:
    def test_is_local_time(self):
        assert localtime.is_local_time(localtime.now())

    def test_is_not_local_time(self):
        dt = datetime.datetime(2016, 8, 5, tzinfo=datetime.timezone.utc)
        assert not localtime.is_local_time(dt)


class TestQuantise:
    @pytest.mark.parametrize(
        ("dt", "timedelta", "rounding_strategy", "result"),
        [
            (
                localtime.datetime(2016, 12, 5, 11, 34, 59),
                datetime.timedelta(minutes=30),
                None,
                localtime.datetime(2016, 12, 5, 11, 30, 0),
            ),
            (
                localtime.datetime(2016, 12, 5, 11, 29, 59),
                datetime.timedelta(minutes=30),
                None,
                localtime.datetime(2016, 12, 5, 11, 30, 0),
            ),
            (
                localtime.datetime(2016, 12, 5, 11, 14, 59),
                datetime.timedelta(minutes=30),
                None,
                localtime.datetime(2016, 12, 5, 11, 0, 0),
            ),
            (
                localtime.datetime(2016, 12, 5, 11, 15, 59),
                datetime.timedelta(minutes=30),
                None,
                localtime.datetime(2016, 12, 5, 11, 30, 0),
            ),
            (
                localtime.datetime(2016, 12, 5, 11, 16, 0),
                datetime.timedelta(minutes=30),
                None,
                localtime.datetime(2016, 12, 5, 11, 30, 0),
            ),
            (
                localtime.datetime(2016, 12, 5, 10, 59, 59),
                datetime.timedelta(hours=2),
                None,
                localtime.datetime(2016, 12, 5, 10, 0, 0),
            ),
            (
                localtime.datetime(2016, 12, 5, 11, 0, 0),
                datetime.timedelta(hours=2),
                None,
                localtime.datetime(2016, 12, 5, 12, 0, 0),
            ),
            (
                localtime.datetime(2016, 12, 31, 13, 34, 59),
                datetime.timedelta(hours=24),
                None,
                localtime.datetime(2017, 1, 1, 0, 0, 0),
            ),
            (
                localtime.datetime(2016, 12, 31, 0, 0, 1),
                datetime.timedelta(days=1),
                None,
                localtime.datetime(2016, 12, 31, 0, 0, 0),
            ),
            # CUSTOM ROUNDING STRATEGIES
            (
                localtime.datetime(2016, 2, 2, 12, 30, 1),
                datetime.timedelta(minutes=30),
                decimal.ROUND_UP,
                localtime.datetime(2016, 2, 2, 13, 0, 0),
            ),
            (
                localtime.datetime(2016, 2, 2, 12, 45, 0),
                datetime.timedelta(minutes=30),
                decimal.ROUND_HALF_UP,
                localtime.datetime(2016, 2, 2, 13, 0, 0),
            ),
            (
                localtime.datetime(2016, 2, 2, 12, 45, 0),
                datetime.timedelta(minutes=30),
                decimal.ROUND_HALF_DOWN,
                localtime.datetime(2016, 2, 2, 12, 30, 0),
            ),
            (
                localtime.datetime(2016, 2, 2, 12, 0, 0),
                datetime.timedelta(days=1),
                decimal.ROUND_HALF_DOWN,
                localtime.datetime(2016, 2, 2, 0, 0, 0),
            ),
            (
                localtime.datetime(2016, 2, 2, 22, 0),
                datetime.timedelta(days=1),
                decimal.ROUND_HALF_UP,
                localtime.datetime(2016, 2, 3, 0, 0, 0),
            ),
            # Test doesn't round beyond limits
            (
                localtime.datetime(2016, 2, 2, 12, 30, 0),
                datetime.timedelta(minutes=30),
                decimal.ROUND_UP,
                localtime.datetime(2016, 2, 2, 12, 30, 0),
            ),
            (
                localtime.datetime(2016, 2, 2, 12, 30, 0),
                datetime.timedelta(minutes=30),
                decimal.ROUND_DOWN,
                localtime.datetime(2016, 2, 2, 12, 30, 0),
            ),
        ],
    )
    def test_quantise(self, dt, timedelta, rounding_strategy, result):
        args = (dt, timedelta)
        kwargs = dict(rounding=rounding_strategy) if rounding_strategy else {}
        assert localtime.quantise(*args, **kwargs) == result


class TestDateBoundaries:
    def test_date_boundaries_for_gmt_date(self):
        date = datetime.date(2017, 1, 1)

        start, end = localtime.date_boundaries(date)

        assert localtime.is_local_time(start)
        assert localtime.is_local_time(end)
        assert (end - start) == datetime.timedelta(days=1)
        assert localtime.datetime(2017, 1, 1) == start
        assert localtime.datetime(2017, 1, 2) == end

    def test_date_boundaries_for_bst_date(self):
        date = datetime.date(2017, 5, 1)

        start, end = localtime.date_boundaries(date)

        assert localtime.is_local_time(start)
        assert localtime.is_local_time(end)
        assert (end - start) == datetime.timedelta(days=1)
        assert localtime.datetime(2017, 5, 1) == start
        assert localtime.datetime(2017, 5, 2) == end

    def test_date_boundaries_for_spring_dst_boundary(self):
        date = datetime.date(2017, 3, 26)

        start, end = localtime.date_boundaries(date)

        assert localtime.is_local_time(start)
        assert localtime.is_local_time(end)
        delta = datetime.timedelta(seconds=(end.timestamp() - start.timestamp()))
        assert delta == datetime.timedelta(seconds=23 * 60 * 60)
        assert localtime.datetime(2017, 3, 26) == start
        assert localtime.datetime(2017, 3, 27) == end

    def test_date_boundaries_for_autumn_dst_boundary(self):
        date = datetime.date(2017, 10, 29)

        start, end = localtime.date_boundaries(date)

        assert localtime.is_local_time(start)
        assert localtime.is_local_time(end)
        delta = datetime.timedelta(seconds=(end.timestamp() - start.timestamp()))
        assert delta == datetime.timedelta(seconds=25 * 60 * 60)
        assert localtime.datetime(2017, 10, 29) == start
        assert localtime.datetime(2017, 10, 30) == end


class TestMonthBoundaries:
    def test_march_2021(self):
        assert localtime.month_boundaries(month=3, year=2021) == (
            factories.local.dt("2021-03-01T00:00"),
            factories.local.dt("2021-04-01T00:00"),
        )


class TestStartOfMonth:
    @pytest.mark.parametrize(
        ("dt", "result"),
        [
            (
                localtime.datetime(2016, 12, 5, 11, 34, 59),
                localtime.datetime(2016, 12, 1, 0, 0, 0),
            ),
            (
                localtime.datetime(2017, 3, 31, 11, 29, 59),
                localtime.datetime(2017, 3, 1, 0, 0, 0),
            ),
        ],
    )
    def test_start_of_month(self, dt, result):
        assert localtime.start_of_month(dt) == result


class TestEndOfMonth:
    @pytest.mark.parametrize(
        ("dt", "result"),
        [
            (
                localtime.datetime(2016, 12, 5, 11, 34, 59),
                localtime.datetime(2017, 1, 1, 0, 0, 0),
            ),
            (
                localtime.datetime(2017, 3, 31, 11, 29, 59),
                localtime.datetime(2017, 4, 1, 0, 0, 0),
            ),
        ],
    )
    def test_end_of_month(self, dt, result):
        assert localtime.end_of_month(dt) == result


class TestAsRange:
    def test_converts_date_to_correct_values(self):
        min_dt, max_dt = localtime.as_range(factories.date("2017-10-01"))

        assert localtime.is_local_time(min_dt)
        assert min_dt == localtime.datetime(2017, 10, 1)

        assert localtime.is_local_time(max_dt)
        assert max_dt == localtime.datetime(2017, 10, 1, 23, 59, 59, 999999)


class TestIsInFutureOrPast:
    def test_with_aware_datetimes(self):
        dt1 = factories.utc.in_the_future(minutes=5), True
        dt2 = factories.utc.in_the_past(minutes=10), False
        dt3 = factories.utc.in_the_future(days=20), True
        dt4 = factories.utc.in_the_past(months=1), False
        for dt, in_future in (dt1, dt2, dt3, dt4):
            assert localtime.is_in_the_future(dt) is in_future
            assert localtime.is_in_the_past(dt) is not in_future

    def test_with_naive_datetimes(self):
        dt = datetime.datetime.now()
        with pytest.raises(ValueError):
            localtime.is_in_the_past(dt)


class TestNextMidnight:
    def test_utc_date(self):
        dt = localtime.next_midnight(factories.date("2017-01-01"))

        assert str(dt.tzinfo) == "Europe/London"
        assert dt == localtime.datetime(2017, 1, 2)

    def test_dst_start(self):
        dt = localtime.next_midnight(factories.date("2018-03-25"))

        assert str(dt.tzinfo) == "Europe/London"
        assert dt == localtime.datetime(2018, 3, 26)

    def test_dst_end(self):
        dt = localtime.next_midnight(factories.date("2018-10-28"))

        assert str(dt.tzinfo) == "Europe/London"
        assert dt == localtime.datetime(2018, 10, 29)

    def test_default_in_different_timezone(self):
        aus_time = zoneinfo.ZoneInfo("Etc/GMT-10")

        with time_machine.travel(
            datetime.datetime(2020, 2, 2, 1, tzinfo=aus_time), tick=False
        ):
            result = localtime.next_midnight(tz=aus_time)

        assert result == datetime.datetime(2020, 2, 3, 0, 0, tzinfo=aus_time)

    @pytest.mark.parametrize(
        "dt,expected",
        [
            # GMT -> BST
            (localtime.datetime(2018, 3, 24), localtime.datetime(2018, 3, 25)),
            (localtime.datetime(2018, 3, 25), localtime.datetime(2018, 3, 26)),
            (localtime.datetime(2018, 3, 26), localtime.datetime(2018, 3, 27)),
            # BST -> GMT
            (localtime.datetime(2018, 10, 27), localtime.datetime(2018, 10, 28)),
            (localtime.datetime(2018, 10, 28), localtime.datetime(2018, 10, 29)),
            (localtime.datetime(2018, 10, 29), localtime.datetime(2018, 10, 30)),
        ],
    )
    def test_dst_end_datetime(self, dt, expected):
        result = localtime.next_midnight(dt)

        assert result == expected
        assert str(result.tzinfo) == "Europe/London"


class TestDaysInThePast:
    def test_is_sane(self):
        assert localtime.days_in_the_past(
            2
        ) == datetime.date.today() - datetime.timedelta(days=2)
        assert localtime.days_in_the_past(
            -20
        ) == datetime.date.today() + datetime.timedelta(days=20)
        assert localtime.days_in_the_past(0) == datetime.date.today()
        assert localtime.days_in_the_past(1) == localtime.yesterday()
        assert localtime.days_in_the_past(-1) == localtime.tomorrow()


class TestDaysInTheFuture:
    def test_is_sane(self):
        assert localtime.days_in_the_future(
            2
        ) == datetime.date.today() + datetime.timedelta(days=2)
        assert localtime.days_in_the_future(
            -20
        ) == datetime.date.today() - datetime.timedelta(days=20)
        assert localtime.days_in_the_future(0) == datetime.date.today()
        assert localtime.days_in_the_future(1) == localtime.tomorrow()
        assert localtime.days_in_the_future(-1) == localtime.yesterday()


class TestLatestDateForDay:
    @pytest.mark.parametrize(
        "start_date,end_date,day_of_month,expected_result",
        (
            ("2017-01-01", "2018-12-31", 9, "2018-12-09"),  # Result in last month.
            ("2017-01-01", "2018-12-08", 9, "2018-11-09"),  # Result in previous month.
            (
                "2017-01-01",
                "2017-03-30",
                31,
                "2017-01-31",
            ),  # Result affected by short month.
            ("2017-01-12", "2017-01-30", 12, "2017-01-12"),  # Result same as from date.
            ("2017-01-12", "2017-01-30", 30, "2017-01-30"),  # Result same as to date.
            ("2017-01-12", "2017-02-10", 11, None),  # Result not in range.
            ("2017-01-01", "2016-01-01", 1, ValueError),  # Invalid range.
            ("2017-01-01", "2018-12-31", 0, ValueError),  # Day too low.
            ("2017-01-01", "2018-12-31", 32, ValueError),  # Day too high.
        ),
    )
    def test_latest_date_for_day(
        self, start_date, end_date, day_of_month, expected_result
    ):
        kwargs = dict(
            start_date=factories.date(start_date),
            end_date=factories.date(end_date),
            day_of_month=day_of_month,
        )
        if isinstance(expected_result, type) and issubclass(expected_result, Exception):
            with pytest.raises(expected_result):
                localtime.latest_date_for_day(**kwargs)
        else:
            if expected_result is not None:
                expected_result = factories.date(expected_result)

            result = localtime.latest_date_for_day(**kwargs)
            assert result == expected_result


class TestIsWithinTheLastYear:
    @pytest.mark.parametrize(
        "now_string, supplied_date_string, within_last_year",
        [
            # Same day should be True:
            ("2019-01-01 00:00:00", "2019-01-01", True),  # Midnight GMT
            ("2019-08-01 00:00:00", "2019-08-01", True),  # Midnight BST
            ("2019-01-01 23:59:59", "2019-01-01", True),  # Just before midnight GMT
            ("2019-08-01 23:59:59", "2019-08-01", True),  # Just after midnight BST
            # Tomorrow should be False:
            ("2019-01-01 00:00:00", "2019-01-02", False),  # Midnight GMT
            ("2019-08-01 00:00:00", "2019-08-02", False),  # Midnight BST
            ("2019-01-01 23:59:59", "2019-01-02", False),  # Just before midnight GMT
            ("2019-08-01 23:59:59", "2019-08-02", False),  # Just after midnight BST
            # Exactly a year ago should be True:
            ("2019-01-01 00:00:00", "2018-01-01", True),  # Midnight GMT
            ("2019-08-01 00:00:00", "2018-08-01", True),  # Midnight BST
            ("2019-01-01 23:59:59", "2018-01-01", True),  # Just before midnight GMT
            ("2019-08-01 23:59:59", "2018-08-01", True),  # Just after midnight BST
            # A year and a day ago should be False:
            ("2019-01-01 00:00:00", "2017-12-31", False),  # Midnight GMT
            ("2019-08-01 00:00:00", "2018-07-31", False),  # Midnight BST
            ("2019-01-01 23:59:59", "2017-12-31", False),  # Just before midnight GMT
            ("2019-08-01 23:59:59", "2018-07-31", False),  # Just after midnight BST
            # Leap years (2020 is a leap year):
            ("2019-02-28 12:00:00", "2018-02-28", True),
            ("2019-02-28 12:00:00", "2018-02-27", False),
            ("2020-02-29 12:00:00", "2019-02-28", True),
            ("2020-02-29 12:00:00", "2019-02-27", False),
            ("2021-02-28 12:00:00", "2020-02-28", True),
            ("2021-02-28 12:00:00", "2020-02-27", False),
        ],
    )
    def test_returns_correct_results_for_dates(
        self, now_string, supplied_date_string, within_last_year
    ):
        now = factories.local.dt(now_string)
        supplied_date = factories.date(supplied_date_string)
        with time_machine.travel(now, tick=False):
            assert localtime.is_within_the_last_year(supplied_date) == within_last_year


class TestWithinLastWeek:
    @pytest.mark.parametrize(
        "now_str, supplied_date_str, is_within_last_year",
        [
            # Same day should be True:
            ("2019-01-01 00:00:00", "2019-01-01", True),  # Midnight
            ("2019-01-01 23:59:59", "2019-01-01", True),  # Just before midnight
            # Tomorrow should be False:
            ("2019-01-01 00:00:00", "2019-01-02", False),  # Midnight
            ("2019-01-01 23:59:59", "2019-01-02", False),  # Just before midnight
            # Exactly a week ago should be True:
            ("2019-01-01 00:00:00", "2018-12-25", True),  # Midnight
            ("2019-01-01 23:59:59", "2018-12-25", True),  # Just before midnight
            # A week and a day ago should be False:
            ("2019-01-01 00:00:00", "2018-12-24", False),  # Midnight
            ("2019-01-01 23:59:59", "2018-12-24", False),  # Just before midnight
        ],
    )
    def test_returns_correct_results_for_dates(
        self, now_str, supplied_date_str, is_within_last_year
    ):
        now = factories.local.dt(now_str)
        supplied_date = factories.date(supplied_date_str)
        with time_machine.travel(now, tick=False):
            assert (
                localtime.is_within_the_last_week(supplied_date) == is_within_last_year
            )


class TestIsDST:
    @pytest.mark.parametrize(
        "naive_datetime,tz,expected",
        (
            # Test London timezone
            (datetime.datetime(2019, 1, 1), zoneinfo.ZoneInfo("Europe/London"), False),
            (datetime.datetime(2019, 6, 1), zoneinfo.ZoneInfo("Europe/London"), True),
            # Test London boundaries
            (
                datetime.datetime(2017, 3, 26, 0, 0),
                zoneinfo.ZoneInfo("Europe/London"),
                False,
            ),
            (
                datetime.datetime(2017, 3, 26, 2, 0),
                zoneinfo.ZoneInfo("Europe/London"),
                True,
            ),
            (
                datetime.datetime(2017, 10, 29, 0, 0),
                zoneinfo.ZoneInfo("Europe/London"),
                True,
            ),
            (
                datetime.datetime(2017, 10, 29, 2, 0),
                zoneinfo.ZoneInfo("Europe/London"),
                False,
            ),
            # UTC should never be DST
            (datetime.datetime(2019, 1, 1), zoneinfo.ZoneInfo("UTC"), False),
            (datetime.datetime(2019, 6, 1), zoneinfo.ZoneInfo("UTC"), False),
            (datetime.datetime(2019, 1, 1), datetime.timezone.utc, False),
            (datetime.datetime(2019, 6, 1), datetime.timezone.utc, False),
            # Test Eastern Australia timezone
            (
                datetime.datetime(2019, 1, 1),
                zoneinfo.ZoneInfo("Australia/Sydney"),
                True,
            ),
            (
                datetime.datetime(2019, 6, 1),
                zoneinfo.ZoneInfo("Australia/Sydney"),
                False,
            ),
            # Test Western Australia timezone (they don't have DST)
            (
                datetime.datetime(2019, 1, 1),
                zoneinfo.ZoneInfo("Australia/Perth"),
                False,
            ),
            (
                datetime.datetime(2019, 6, 1),
                zoneinfo.ZoneInfo("Australia/Perth"),
                False,
            ),
        ),
    )
    def test_returns_correct_values(self, naive_datetime, tz, expected):
        local_dt = timezone.make_aware(naive_datetime, timezone=tz)
        assert localtime.is_dst(local_dt) == expected

    def test_raises_value_error_for_naive_dt(self):
        with pytest.raises(ValueError):
            localtime.is_dst(datetime.datetime(2008, 4, 5))


class TestIsLocaltimeMidnight:
    @pytest.mark.parametrize(
        "dt",
        [
            # Valid datetimes in UTC and local, summer and winter.
            factories.local.dt("2019-12-01 00:00"),
            factories.utc.dt("2019-12-01 00:00"),
            factories.local.dt("2019-06-01 00:00"),
            factories.utc.dt("2019-06-01 23:00"),
        ],
    )
    def test_returns_true_for_localtime_midnights(self, dt):
        assert localtime.is_localtime_midnight(dt)

    @pytest.mark.parametrize(
        "dt",
        [
            factories.local.dt("2019-12-01 12:01"),
            factories.utc.dt("2019-12-01 12:01"),
            factories.local.dt("2019-06-01 12:01"),
            factories.utc.dt("2019-06-01 00:00"),
        ],
    )
    def test_returns_false_for_non_localtime_midnights(self, dt):
        assert not localtime.is_localtime_midnight(dt)

    def test_returns_false_if_timezone_differs(self):
        assert not localtime.is_localtime_midnight(
            factories.utc.dt("2020-12-01 00:00"),
            tz=zoneinfo.ZoneInfo("Europe/Paris"),
        )


class TestCombine:
    @pytest.mark.parametrize(
        "date, time, tz_name, expected",
        (
            (
                factories.date("1 Jan 2020"),
                factories.time("00:00"),
                "UTC",
                datetime.datetime(2020, 1, 1, tzinfo=localtime.UTC),
            ),
            (
                factories.date("1 Jan 2020"),
                factories.time("00:00"),
                "UTC",
                datetime.datetime(2020, 1, 1, tzinfo=datetime.timezone.utc),
            ),
            (
                factories.date("1 Jun 2020"),
                factories.time("01:00"),
                "Europe/London",
                datetime.datetime(2020, 6, 1, 1, 0).astimezone(
                    zoneinfo.ZoneInfo("Europe/London")
                ),
            ),
            (
                factories.date("1 Jul 2021"),
                factories.time("02:30"),
                "Europe/London",
                datetime.datetime(2021, 7, 1, 2, 30).astimezone(
                    zoneinfo.ZoneInfo("Europe/London")
                ),
            ),
        ),
    )
    def test_combines_as_expected(self, date, time, tz_name, expected):
        tz = zoneinfo.ZoneInfo(tz_name)
        assert localtime.combine(date, time, tz) == expected


class TestNextDateWithDayOfMonth:
    @pytest.mark.parametrize(
        "current_date, day_of_month, expected",
        (
            (factories.date("1 Jan 2020"), 1, factories.date("1 Feb 2020")),
            (factories.date("28 Jan 2020"), 1, factories.date("1 Feb 2020")),
            (factories.date("31 Jan 2020"), 31, factories.date("29 Feb 2020")),
            (factories.date("31 Dec 2019"), 2, factories.date("2 Jan 2020")),
        ),
    )
    def test_next_date_with_day_of_month(self, current_date, day_of_month, expected):
        assert (
            localtime.next_date_with_day_of_month(
                date=current_date, day_of_month=day_of_month
            )
            == expected
        )

    def test_raises_if_datetime_is_used(self):
        date_time = factories.utc.dt("1 Jan 2020 15:00")

        with pytest.raises(TypeError):
            localtime.next_date_with_day_of_month(date_time, 1)


class TestConsolidateIntoIntervals:
    def test_invalid_input(self):
        with pytest.raises(ValueError):
            localtime.consolidate_into_intervals([])

    @pytest.mark.parametrize(
        "dates_passed,dates_expected",
        [
            (
                [localtime.today()],
                [(localtime.today(), localtime.today())],
            ),
            (
                [
                    localtime.today(),
                    localtime.yesterday(),
                    localtime.days_in_the_future(2),
                ],
                [
                    (localtime.yesterday(), localtime.today()),
                    (localtime.days_in_the_future(2), localtime.days_in_the_future(2)),
                ],
            ),
        ],
    )
    def test_valid_input(self, dates_passed, dates_expected):
        assert localtime.consolidate_into_intervals(dates_passed) == dates_expected


class TestDatetimeFromUTCUnixTimestamp:
    @override_settings(TIME_ZONE="Europe/London")
    def test_timestamp_in_london_timezone(self):
        timestamp = 1605804900
        dt = localtime.datetime_from_epoch_timestamp(timestamp)
        assert dt == localtime.datetime(2020, 11, 19, 16, 55)

    @override_settings(TIME_ZONE="Etc/GMT-10")
    def test_timestamp_in_aus_timezone(self):
        timestamp = 1605804900
        dt = localtime.datetime_from_epoch_timestamp(timestamp)
        assert dt == localtime.datetime(2020, 11, 20, 2, 55)

    def test_timestamp_with_timestamp_argument(self):
        aus_time = zoneinfo.ZoneInfo("Etc/GMT-10")
        timestamp = 1605804900

        dt = localtime.datetime_from_epoch_timestamp(timestamp, tz=aus_time)
        assert dt == datetime.datetime(2020, 11, 20, 2, 55, tzinfo=aus_time)

    @override_settings(TIME_ZONE="Europe/London")
    def test_timestamp_british_summer_time_before_clocks_move_forward(self):
        """
        On the 29th of March 2020 at 1:00 am Europe/London = 1:00 am UTC
        the clocks move forward by an hour.
        That means that the clocks read 2:00 am at that point.
        """
        # Before clocks move forward
        # 29th of March 2020 0:30am UTC = 0:30am Europe/London
        timestamp = datetime.datetime(
            2020, 3, 29, 0, 30, tzinfo=localtime.UTC
        ).timestamp()

        dt = localtime.datetime_from_epoch_timestamp(timestamp)

        # Before daylight savings the time should be 0:30am (same as UTC)
        assert not localtime.is_dst(dt)
        assert dt.year == 2020
        assert dt.month == 3
        assert dt.day == 29
        assert dt.hour == 0
        assert dt.minute == 30

    @override_settings(TIME_ZONE="Europe/London")
    def test_timestamp_british_summer_time_after_clocks_move_forward(self):
        """
        On the 29th of March 2020 at 1:00 am Europe/London = 1:00 am UTC
        the clocks move forward by an hour.
        That means that the clocks read 2:00 am at that point.
        """
        # After clocks move forward
        # 29th of March 2020 1:30am UTC = 2:30 am Europe/London
        timestamp = datetime.datetime(
            2020, 3, 29, 1, 30, tzinfo=localtime.UTC
        ).timestamp()

        dt = localtime.datetime_from_epoch_timestamp(timestamp)

        # During daylight savings the time should be 2:30am (one hour forward)
        assert localtime.is_dst(dt)
        assert dt.year == 2020
        assert dt.month == 3
        assert dt.day == 29
        assert dt.hour == 2
        assert dt.minute == 30

    @override_settings(TIME_ZONE="Europe/London")
    def test_timestamp_british_summer_time_before_clocks_move_backward(self):
        """
        On the 25th of October 2020 at 2:00 am Europe/London = 1:00 am UTC
        the clocks move backwards by an hour.
        That means that the clocks read 1:00 am at that point.
        """
        # Before clocks move backwards
        # 25th of October 2020 0:30am UTC = 1:30am Europe/London
        timestamp = datetime.datetime(
            2020, 10, 25, 0, 30, tzinfo=localtime.UTC
        ).timestamp()

        dt = localtime.datetime_from_epoch_timestamp(timestamp)

        # During daylight savings the time should be 1:30am (one hour forward)
        assert localtime.is_dst(dt)
        assert dt.year == 2020
        assert dt.month == 10
        assert dt.day == 25
        assert dt.hour == 1
        assert dt.minute == 30

    @override_settings(TIME_ZONE="Europe/London")
    def test_timestamp_british_summer_time_after_clocks_move_backward(self):
        """
        On the 25th of October 2020 at 2:00 am Europe/London = 1:00 am UTC
        the clocks move backwards by an hour.
        That means that the clocks read 1:00 am at that point.
        """
        # After clocks move backwards
        # 25th of October 2020 1:30am UTC = 1:30am Europe/London
        timestamp = datetime.datetime(
            2020, 10, 25, 1, 30, tzinfo=localtime.UTC
        ).timestamp()

        dt = localtime.datetime_from_epoch_timestamp(timestamp)

        # After daylight savings ends the time should be 1:30am (same as UTC)
        assert not localtime.is_dst(dt)
        assert dt.year == 2020
        assert dt.month == 10
        assert dt.day == 25
        assert dt.hour == 1
        assert dt.minute == 30


class TestPeriodExceedsOneYear:
    @pytest.mark.parametrize(
        ("period_start_at", "first_dt_exceeding_one_year"),
        [
            (
                # Basic case.
                localtime.datetime(2021, 1, 1),
                localtime.datetime(2022, 1, 1, microsecond=1),
            ),
            (
                # A leap year.
                localtime.datetime(2020, 1, 1),
                localtime.datetime(2021, 1, 1, microsecond=1),
            ),
            (
                # Start on a leap year, Feb 28th.
                localtime.datetime(2020, 2, 28),
                localtime.datetime(2021, 2, 28, microsecond=1),
            ),
            (
                # Start on a leap year, Feb 29th.
                localtime.datetime(2020, 2, 29),
                localtime.datetime(2021, 3, 1, microsecond=1),  # !important
            ),
            (
                # Start on a leap year, March 1st.
                localtime.datetime(2020, 3, 1),
                localtime.datetime(2021, 3, 1, microsecond=1),
            ),
            (
                # End on a leap year, Feb 28th.
                localtime.datetime(2019, 2, 28),
                localtime.datetime(2020, 2, 28, microsecond=1),
            ),
            (
                # End on a leap year, March 1st.
                localtime.datetime(2019, 3, 1),
                localtime.datetime(2020, 3, 1, microsecond=1),
            ),
            (
                # Clock moves backward twice.
                localtime.datetime(2021, 10, 31),
                localtime.datetime(2022, 10, 31, microsecond=1),
            ),
            (
                # Clock moves forward twice.
                localtime.datetime(2021, 3, 28),
                localtime.datetime(2022, 3, 28, microsecond=1),
            ),
        ],
    )
    def test_period_exceeds_one_year(
        self, period_start_at, first_dt_exceeding_one_year
    ):
        assert localtime.period_exceeds_one_year(
            period_start_at, first_dt_exceeding_one_year
        )
        assert not localtime.period_exceeds_one_year(
            period_start_at,
            first_dt_exceeding_one_year - relativedelta.relativedelta(microseconds=1),
        )


class TestParseDate:
    def test_returns_date(self):
        assert localtime.parse_date("2020-01-01") == datetime.date(2020, 1, 1)

    def test_errors_if_invalid(self):
        with pytest.raises(ValueError) as exc_info:
            localtime.parse_date("abcd")

        assert "Invalid isoformat string" in str(exc_info.value)


class TestParseDatetime:
    @override_settings(TIME_ZONE="Australia/Sydney")
    def test_returns_datetime(self):
        assert localtime.parse_dt("2020-01-01 10:11:12") == datetime.datetime(
            2020, 1, 1, 10, 11, 12, tzinfo=zoneinfo.ZoneInfo("Australia/Sydney")
        )

    @override_settings(TIME_ZONE="Australia/Sydney")
    def test_assumes_midnight(self):
        assert localtime.parse_dt("2020-01-01") == datetime.datetime(
            2020, 1, 1, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("Australia/Sydney")
        )

    def test_errors_if_timezone_specified(self):
        with pytest.raises(ValueError) as exc_info:
            localtime.parse_dt("2020-01-01 10:11 +01:00")

        assert "expects a naive datetime" in str(exc_info.value)

    def test_errors_if_invalid(self):
        with pytest.raises(ValueError) as exc_info:
            localtime.parse_date("abcd")

        assert "Invalid isoformat string" in str(exc_info.value)


class TestStrftime:
    @override_settings(TIME_ZONE="Europe/Berlin")
    def test_formats_datetime_in_local_timezone(self):
        dt = datetime.datetime(2023, 10, 1, 22, 30, 0, tzinfo=datetime.timezone.utc)
        fmt = "%Y-%m-%d %H:%M:%S %z"

        assert localtime.strftime(dt, fmt) == "2023-10-02 00:30:00 +0200"
        assert (
            localtime.strftime(dt, fmt, tz=zoneinfo.ZoneInfo("Europe/London"))
            == "2023-10-01 23:30:00 +0100"
        )



from __future__ import annotations

from typing import Any, Protocol

import structlog
from django import http
from django.conf import settings


logger = structlog.get_logger("events")

__all__ = ["publish"]

METADATA = {
    "release": "GIT_SHA",
    "aws_private_ip": "AWS_LOCAL_IP",
    "aws_instance_id": "AWS_INSTANCE_ID",
    "aws_availability_zone": "AWS_AVAILABILITY_ZONE",
    "aws_auto_scaling_group": "AWS_AUTO_SCALING_GROUP",
}


class Account(Protocol):
    number: str


def publish(
    event: str,
    params: dict[str, Any] | None = None,
    meta: dict[str, Any] | None = None,
    account: Account | None = None,
    request: http.HttpRequest | None = None,
) -> None:
    """
    Publish an event.

    - `params` are values that were used to create the event (eg the path of a
      request)
    - `meta` are contextual values around the event (eg the IP address of the
      person making the request)

    Note, structlog will add a timestamp.
    """
    payload: dict[str, Any] = {"event": event}
    if params is not None:
        payload["params"] = params
    if meta is None:
        meta = {}

    # If the event relates to a single account, we include its number so it's easy to filter down
    # to events that affect one account in Loggly.
    if account is not None:
        payload["account"] = account.number

    # Add static metadata from settings.
    for key, setting in METADATA.items():
        value = getattr(settings, setting, "")
        if value:
            meta[key] = value

    # Add metadata from request
    if request is not None:
        meta.update(_request_meta(request))

    payload["meta"] = meta

    _log(payload)


def _log(event: dict[str, Any]) -> None:
    """
    Log the event.
    """
    event_ = event.copy()
    name = event_.pop("event")
    logger.info(name, **event_)


def _request_meta(request: http.HttpRequest) -> dict[str, Any]:
    """
    Extract relevant meta information from a request instance.
    """
    meta = {"ip_address": request.META["REMOTE_ADDR"]}
    if "HTTP_USER_AGENT" in request.META:
        meta["user_agent"] = request.META["HTTP_USER_AGENT"]
    if hasattr(request, "session") and request.session.session_key:
        meta["session"] = request.session.session_key
    return meta
